import { Function, Supplier, BiConsumer, BiFunction, Comparator, Predicate } from "../functions";
import { MutableString, MutableNumber, Holder, NumberSummaryStatistics } from './mutableCollections';
import { Map } from '../map';
import Optional from "../optional";
/**
 * A mutable reduction operation that accumulates input elements into a mutable result container,
 * optionally transforming the accumulated result into a final representation after all input elements have been processed.
 *
 * Examples of mutable reduction operations include: accumulating elements into an array; concatenating strings;
 * computing summary information about elements such as sum, min, max, or average;
 * computing "pivot table" summaries such as "maximum valued transaction by seller", etc.
 *
 * Collectors provides implementations of many common mutable reductions.
 *
 * A Collector is specified by four functions that work together to accumulate entries into a mutable result container,
 * and optionally perform a final transform on the result. They are:
 * creation of a new result container (supplier())
 * incorporating a new data element into a result container (accumulator())
 * combining two result containers into one (combiner())
 * performing an optional final transform on the container (finisher())
 */
export declare class Collector<T, A, R> {
    private supp;
    private accu;
    private comb;
    private fini;
    private constructor();
    supplier: () => Supplier<A>;
    accumulator: () => BiConsumer<A, T>;
    combiner: () => BiFunction<A>;
    finisher: () => Function<A, R>;
    /**
    * A Collector is specified by four functions that work together to accumulate entries into a mutable result container,
    * and optionally perform a final transform on the result. They are:
    * @param supplier function that returns of a new result container
    * @param accumulator function for incorporating a new data element into a result container
    * @param combiner function for combining two result containers into one
    * @param finisher function for performing an optional final transform on the container
    */
    static of<T, A, R>(supplier: Supplier<A>, accumulator: BiConsumer<A, T>, combiner: BiFunction<A>, finisher: Function<A, R>): Collector<T, A, R>;
}
/**
 * defines many useful collectors for easily reducing/collecting data together
 */
declare class Collectors {
    /**
     * A collector that combines input into an array.
     */
    static toArray<T>(): Collector<T, T[], T[]>;
    /**
     * alias of toArray()
     */
    static toList<T>(): Collector<T, T[], T[]>;
    /**
     * Returns a Collector that concatenates the input elements into a String, in encounter order.
     */
    static joining(): Collector<string, MutableString, string>;
    /**
     * Returns a Collector that concatenates the input elements, separated by the specified delimiter, in encounter order.
     * @param delimiter - the delimiter to be used between each element
     */
    static joining(delimiter: string): Collector<string, MutableString, string>;
    /**
     * Returns a Collector that concatenates the input elements, separated by the specified delimiter, with the specified prefix and suffix, in encounter order.
     * @param delimiter - the delimiter to be used between each element
     * @param prefix - the sequence of characters to be used at the beginning of the joined result
     * @param suffix - the sequence of characters to be used at the end of the joined result
     */
    static joining(delimiter: string, prefix: string, suffix: string): Collector<string, MutableString, string>;
    /**
     * Returns a Collector that produces the arithmetic mean of a number-valued function applied to the input elements.
     * @param mapper function to transform input elements into a number
     */
    static averagingNumber<I>(mapper: Function<I, number>): Collector<I, MutableNumber, number>;
    /**
     * returns a Collector that produces the arithmetic mean of input numbers
     */
    static averaging(): Collector<number, MutableNumber, number>;
    /**
     * Adapts a Collector to perform an additional finishing transformation.
     * @param downstream - a collector
     * @param finisher - a function to be applied to the final result of the downstream collector
     */
    static collectingAndThen<I, M, A, O>(downStream: Collector<I, M, A>, finisher: Function<A, O>): Collector<I, M, O>;
    /**
     * Returns a Collector accepting elements of type I that counts the number of input elements. If no elements are present, the result is 0.
     */
    static counting<I>(): Collector<I, MutableNumber, number>;
    /**
     * Returns a Collector implementing a "group by" operation on input elements of type T,
     * grouping elements according to a classification function, and returning the results in a Map.
     * The classification function maps elements to some key type K. The collector produces a
     * Map<K, T[]> whose keys are the values resulting from applying the classification function to the input elements,
     * and whose corresponding values are arrays containing the input elements which map to the associated key under the classification function.
     * @param classifier - the classifier function mapping input elements to keys
     */
    static groupingBy<T, K>(classifier: Function<T, K>): Collector<T, Map<K, T[]>, Map<K, T[]>>;
    /**
     * Returns a Collector implementing a cascaded "group by" operation on input elements of type T,
     * grouping elements according to a classification function, and then performing a reduction operation
     * on the values associated with a given key using the specified downstream Collector.
     * The classification function maps elements to some key type K. The downstream collector operates
     * on elements of type T and produces a result of type D. The resulting collector produces a Map<K, D>.
     * @param classifier - a classifier function mapping input elements to keys
     * @param downstream - a Collector implementing the downstream reduction
     */
    static groupingBy<T, K, A, D>(classifier: Function<T, K>, downstream: Collector<T, A, D>): Collector<T, Map<K, T[]>, Map<K, D>>;
    /**
     * Adapts a Collector accepting elements of type U to one accepting elements of type T by applying a mapping function to each input element before accumulation.
     * @param mapper - a function to be applied to the input elements
     * @param downstream - a collector which will accept mapped values
     */
    static mapping<I, II, A, R>(mapper: Function<I, II>, downstream: Collector<II, A, R>): Collector<I, A, R>;
    /**
     * Returns a Collector that will return the max value as determined by the default comparator
     * See, Comparator.default(); the max value is returned in an optional, or empty if no value
     * was found.
     */
    static maxBy<I>(): Collector<I, Holder<I>, Optional<I>>;
    /**
     * Returns a Collector that will return the max value as determined by the given comparator
     * the max value is returned in an optional, or empty if no value was found.
     * @param comparator - a Comparator for comparing elements
     */
    static maxBy<I>(comparator: Comparator<I>): Collector<I, Holder<I>, Optional<I>>;
    /**
     * Returns a Collector that will return the min value as determined by the default comparator
     * See, Comparator.default(); the min value is returned in an optional, or empty if no value
     * was found.
     */
    static minBy<I>(): Collector<I, Holder<I>, Optional<I>>;
    /**
     * Returns a Collector that will return the min value as determined by the given comparator
     * the min value is returned in an optional, or empty if no value was found.
     * @param comparator - a Comparator for comparing elements
     */
    static minBy<I>(comparator: Comparator<I>): Collector<I, Holder<I>, Optional<I>>;
    /**
     * Returns a Collector which partitions the input elements according to a Predicate, and organizes them into a Map<Boolean, T[]>.
     * @param predicate - a predicate used for classifying input elements
     */
    static partitioningBy<T, A, D>(predicate: Predicate<T>): Collector<T, Map<boolean, T[]>, Map<boolean, T[]>>;
    /**
     * Returns a Collector which partitions the input elements according to a Predicate, reduces the values in each partition according to another Collector, and organizes them into a Map<Boolean, D>
     * whose values are the result of the downstream reduction.
     * @param predicate - a predicate used for classifying input elements
     * @param downstream - a Collector implementing the downstream reduction
     */
    static partitioningBy<T, A, D>(predicate: Predicate<T>, downStream: Collector<T, A, D>): Collector<T, Map<boolean, T[]>, Map<boolean, D>>;
    /**
    * Returns a Collector which performs a reduction of its input elements under a specified BinaryOperator. The result is described as an Optional<T>.
    * @param reducer - a BinaryOperator<T> used to reduce the input elements
    */
    static reducing<I>(reducer: BiFunction<I>): Collector<I, I[], Optional<I>>;
    /**
     * Returns a Collector which performs a reduction of its input elements under a specified BinaryOperator using the provided identity.
     * @param reducer - a BinaryOperator<T> used to reduce the input elements
     * @param identity - the identity value for the reduction (also, the value that is returned when there are no input elements)
     */
    static reducing<I>(reducer: BiFunction<I>, identity: I): Collector<I, I[], Optional<I>>;
    /**
     * Returns a Collector which performs a reduction of its input elements under a specified BinaryOperator using the provided identity.
     * @param identity - the identity value for the reduction (also, the value that is returned when there are no input elements)
     * @param mapper - a mapping function to apply to each input value
     * @param reducer - a BinaryOperator<U> used to reduce the mapped values
     */
    static reducing<I, U>(reducer: BiFunction<U>, identity: U, mapper: Function<I, U>): Collector<I, I[], Optional<U>>;
    /**
     * Returns a Collector that retuns summary statistics for the resulting values
     */
    static summarizingNumber(): Collector<number, NumberSummaryStatistics, NumberSummaryStatistics>;
    /**
     * Returns a Collector which applies an number- mapping function to each input element, and returns summary statistics for the resulting values.
     * @param numberMapper - mapping function that returns a number value
     */
    static summarizingNumber<T>(numberMapper: Function<T, number>): Collector<T, NumberSummaryStatistics, NumberSummaryStatistics>;
    /**
     * Returns a Collector that produces the sum of numbers. If no elements are present, the result is 0.
     */
    static summingNumber(): Collector<number, MutableNumber, number>;
    /**
     * Returns a Collector that produces the sum of a number-valued function applied to the input elements. If no elements are present, the result is 0.
     * @param numberMapper - a function to transform input elements to a number
     */
    static summingNumber<T>(numberMapper: Function<T, number>): Collector<T, MutableNumber, number>;
    /**
     * Returns a Collector that accumulates elements into a Map whose keys and values are the result of
     * applying the provided mapping functions to the input elements.
     * If the mapped keys contains duplicates (according to Object.equals(Object)),
     * an IllegalStateException is thrown when the collection operation is performed.
     * If the mapped keys may have duplicates, use toMap(Function, Function, BiFunction) instead
     * @param keyMapper - a mapping function to produce keys
     * @param valueMapper - a mapping function to produce values
     */
    static toMap<I, K, V>(keyMapper: Function<I, K>, valueMapper: Function<I, V>): Collector<I, Map<K, V>, Map<K, V>>;
    /**
     * Returns a Collector that accumulates elements into a Map whose keys and values are the result of applying
     * the provided mapping functions to the input elements. If the mapped keys contains duplicates,
     * the value mapping function is applied to each equal element, and the results are merged using the provided merging function.
     * @param keyMapper - a mapping function to produce keys
     * @param valueMapper - a mapping function to produce values
     * @param merger - a merge function, used to resolve collisions between values associated with the same key, as supplied to Map.merge(Object, Object, BiFunction)
     */
    static toMap<I, K, V>(keyMapper: Function<I, K>, valueMapper: Function<I, V>, merger: BiFunction<V>): Collector<I, Map<K, V>, Map<K, V>>;
}
export default Collectors;
