"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var functions_1 = require("../functions");
var mutableCollections_1 = require("./mutableCollections");
var map_1 = require("../map");
var stream_1 = require("../stream");
var errors_1 = require("../errors");
/**
 * A mutable reduction operation that accumulates input elements into a mutable result container,
 * optionally transforming the accumulated result into a final representation after all input elements have been processed.
 *
 * Examples of mutable reduction operations include: accumulating elements into an array; concatenating strings;
 * computing summary information about elements such as sum, min, max, or average;
 * computing "pivot table" summaries such as "maximum valued transaction by seller", etc.
 *
 * Collectors provides implementations of many common mutable reductions.
 *
 * A Collector is specified by four functions that work together to accumulate entries into a mutable result container,
 * and optionally perform a final transform on the result. They are:
 * creation of a new result container (supplier())
 * incorporating a new data element into a result container (accumulator())
 * combining two result containers into one (combiner())
 * performing an optional final transform on the container (finisher())
 */
var Collector = /** @class */ (function () {
    function Collector(supp, accu, comb, fini) {
        var _this = this;
        this.supplier = function () { return _this.supp; };
        this.accumulator = function () { return _this.accu; };
        this.combiner = function () { return _this.comb; };
        this.finisher = function () { return _this.fini; };
        this.supp = supp;
        this.accu = accu;
        this.comb = comb;
        this.fini = fini;
    }
    /**
    * A Collector is specified by four functions that work together to accumulate entries into a mutable result container,
    * and optionally perform a final transform on the result. They are:
    * @param supplier function that returns of a new result container
    * @param accumulator function for incorporating a new data element into a result container
    * @param combiner function for combining two result containers into one
    * @param finisher function for performing an optional final transform on the container
    */
    Collector.of = function (supplier, accumulator, combiner, finisher) {
        return new Collector(supplier, accumulator, combiner, finisher);
    };
    return Collector;
}());
exports.Collector = Collector;
;
/**
 * defines many useful collectors for easily reducing/collecting data together
 */
var Collectors = /** @class */ (function () {
    function Collectors() {
    }
    /**
     * A collector that combines input into an array.
     */
    Collectors.toArray = function () {
        var supplier = function () { return []; };
        var accumulator = function (list, item) { return list.push(item); };
        var combiner = function (list1, list2) { return list1.concat(list2); };
        var finisher = functions_1.Function.identity();
        return Collector.of(supplier, accumulator, combiner, finisher);
    };
    /**
     * alias of toArray()
     */
    Collectors.toList = function () {
        return Collectors.toArray();
    };
    Collectors.joining = function (delimiter, prefix, suffix) {
        var ifElseBlank = function (val) { return val ? val : ""; };
        var delimiterToUse = ifElseBlank(delimiter);
        var supplier = function () { return mutableCollections_1.MutableString.empty(); };
        var accumulator = function (mutable, str) { return mutable.append(str + delimiterToUse); };
        var combiner = function (mutable1, mutable2) { return mutable1.concat(mutable2); };
        var finisher = function (mutable) {
            var valueWithExtraDelimiter = mutable.getValue();
            var valueWithOutExtraDelimiter = delimiterToUse.length === 0
                ? valueWithExtraDelimiter
                : valueWithExtraDelimiter.slice(0, -delimiterToUse.length);
            return ifElseBlank(prefix) + valueWithOutExtraDelimiter + ifElseBlank(suffix);
        };
        return Collector.of(supplier, accumulator, combiner, finisher);
    };
    /**
     * Returns a Collector that produces the arithmetic mean of a number-valued function applied to the input elements.
     * @param mapper function to transform input elements into a number
     */
    Collectors.averagingNumber = function (mapper) {
        var supplier = mutableCollections_1.MutableNumber.empty;
        var accumulator = function (mutable, item) { return mutable.add(mapper(item)); };
        var combiner = function (mNum1, mNum2) { return mNum1.addTogether(mNum2); };
        var finisher = function (mutable) {
            return mutable.getInputCount() > 0
                ? mutable.getTotal() / mutable.getInputCount()
                : 0;
        };
        return Collector.of(supplier, accumulator, combiner, finisher);
    };
    /**
     * returns a Collector that produces the arithmetic mean of input numbers
     */
    Collectors.averaging = function () {
        return Collectors.averagingNumber(functions_1.Function.identity());
    };
    /**
     * Adapts a Collector to perform an additional finishing transformation.
     * @param downstream - a collector
     * @param finisher - a function to be applied to the final result of the downstream collector
     */
    Collectors.collectingAndThen = function (downStream, finisher) {
        var newFinisher = function (input) { return finisher(downStream.finisher()(input)); };
        return Collector.of(downStream.supplier(), downStream.accumulator(), downStream.combiner(), newFinisher);
    };
    /**
     * Returns a Collector accepting elements of type I that counts the number of input elements. If no elements are present, the result is 0.
     */
    Collectors.counting = function () {
        var supplier = mutableCollections_1.MutableNumber.empty;
        var accumulator = function (mutable, item) { return mutable.add(1); };
        var combiner = function (mNum1, mNum2) { return mNum1.addTogether(mNum2); };
        var finisher = function (mutable) { return mutable.getInputCount(); };
        return Collector.of(supplier, accumulator, combiner, finisher);
    };
    Collectors.groupingBy = function (classifier, downstream) {
        var supplier = function () { return map_1.Map.empty(); };
        var accumulator = function (map, item) { return map.merge(classifier(item), [item], function (l1, l2) { return l1.concat(l2); }); };
        var combiner = function (map1, map2) {
            map1.putAll(map2);
            return map1;
        };
        if (downstream) {
            var finisher = function (initial) {
                var newMap = map_1.Map.empty();
                initial.forEach(function (key, valList) {
                    return newMap.put(key, stream_1.default.of(valList).collect(downstream));
                });
                return newMap;
            };
            return Collector.of(supplier, accumulator, combiner, finisher);
        }
        else {
            return Collector.of(supplier, accumulator, combiner, functions_1.Function.identity());
        }
    };
    /**
     * Adapts a Collector accepting elements of type U to one accepting elements of type T by applying a mapping function to each input element before accumulation.
     * @param mapper - a function to be applied to the input elements
     * @param downstream - a collector which will accept mapped values
     */
    Collectors.mapping = function (mapper, downstream) {
        return Collector.of(downstream.supplier(), function (mutable, item) { return downstream.accumulator()(mutable, mapper(item)); }, downstream.combiner(), downstream.finisher());
    };
    Collectors.maxBy = function (comparator) {
        var comparatorToUse = comparator ? comparator : functions_1.Comparator.default();
        var supplier = function () { return new mutableCollections_1.Holder(); };
        var accumulator = function (mutable, item) {
            var currentMax = mutable.get();
            if (currentMax.isPresent()) {
                var newMax = currentMax
                    .map(function (current) { return returnLargest(current, item, comparatorToUse); })
                    .get();
                mutable.set(newMax);
            }
            else {
                mutable.set(item);
            }
        };
        var combiner = function (h1, h2) {
            var first = h1.get();
            var second = h2.get();
            if (first.isPresent() && second.isPresent()) {
                return new mutableCollections_1.Holder(returnLargest(first.get(), second.get(), comparatorToUse));
            }
            else if (first.isPresent()) {
                return h1;
            }
            else {
                return h2;
            }
        };
        var finisher = function (mutable) { return mutable.get(); };
        return Collector.of(supplier, accumulator, combiner, finisher);
    };
    Collectors.minBy = function (comparator) {
        var comparatorToUse = comparator ? comparator : functions_1.Comparator.default();
        var supplier = function () { return new mutableCollections_1.Holder(); };
        var accumulator = function (mutable, item) {
            var currentMin = mutable.get();
            if (currentMin.isPresent()) {
                var newMin = currentMin
                    .map(function (current) { return returnSmallest(current, item, comparatorToUse); })
                    .get();
                mutable.set(newMin);
            }
            else {
                mutable.set(item);
            }
        };
        var combiner = function (h1, h2) {
            var first = h1.get();
            var second = h2.get();
            if (first.isPresent() && second.isPresent()) {
                return new mutableCollections_1.Holder(returnSmallest(first.get(), second.get(), comparatorToUse));
            }
            else if (first.isPresent()) {
                return h1;
            }
            else {
                return h2;
            }
        };
        var finisher = function (mutable) { return mutable.get(); };
        return Collector.of(supplier, accumulator, combiner, finisher);
    };
    Collectors.partitioningBy = function (predicate, downStream) {
        var supplier = function () { return map_1.Map.of(true, [], false, []); };
        var accumulator = function (map, item) { return map.merge(predicate(item), [item], function (l1, l2) { return l1.concat(l2); }); };
        var combiner = function (map1, map2) {
            map1.putAll(map2);
            return map1;
        };
        if (downStream) {
            var Function_1 = function (map) {
                var maps = map_1.Map.of(true, stream_1.default.of(map.getOptional(true).orElse([]))
                    .collect(downStream), false, stream_1.default.of(map.getOptional(false).orElse([]))
                    .collect(downStream));
                return maps;
            };
            return Collector.of(supplier, accumulator, combiner, Function_1);
        }
        else {
            return Collector.of(supplier, accumulator, combiner, functions_1.Function.identity());
        }
    };
    Collectors.reducing = function (reducer, identity, mapper) {
        var supplier = function () { return []; };
        var accumulator = function (list, item) { return list.push(item); };
        var combiner = function (list1, list2) { return list1.concat(list2); };
        if (mapper) {
            var finisher = function (iList) { return stream_1.default.of(iList)
                .map(mapper)
                .reduce(reducer, identity); };
            return Collector.of(supplier, accumulator, combiner, finisher);
        }
        else {
            var finisher = function (iList) { return stream_1.default.of(iList)
                .reduce(reducer, identity); };
            return Collector.of(supplier, accumulator, combiner, finisher);
        }
    };
    Collectors.summarizingNumber = function (numberMapper) {
        var isT = function (item) { return numberMapper ? true : false; };
        var mapper = function (item) {
            if (isT(item)) {
                var i = item;
                if (numberMapper) {
                    return numberMapper(i);
                }
            }
            else {
                return item;
            }
            throw new Error("got an unexpected type");
        };
        var supplier = mutableCollections_1.NumberSummaryStatistics.create;
        var accumulator = function (mutable, item) {
            return mutable.accept(mapper(item));
        };
        var combiner = function (mNum1, mNum2) {
            mNum1.combine(mNum2);
            return mNum1;
        };
        return Collector.of(supplier, accumulator, combiner, functions_1.Function.identity());
    };
    Collectors.summingNumber = function (numberMapper) {
        var isT = function (item) { return numberMapper ? true : false; };
        var mapper = function (item) {
            if (isT(item)) {
                var i = item;
                if (numberMapper) {
                    return numberMapper(i);
                }
            }
            else {
                return item;
            }
            throw new Error("got an unexpected type");
        };
        var supplier = mutableCollections_1.MutableNumber.empty;
        var accumulator = function (mutable, num) { return mutable.add(mapper(num)); };
        var combiner = function (m1, m2) { return m1.addTogether(m2); };
        var finisher = function (mut) { return mut.getTotal(); };
        return Collector.of(supplier, accumulator, combiner, finisher);
    };
    Collectors.toMap = function (keyMapper, valueMapper, merger) {
        var supplier = map_1.Map.empty;
        var accumulator = function (map, input) {
            var key = keyMapper(input);
            var value = valueMapper(input);
            if (map.containsKey(key)) {
                if (merger) {
                    map.merge(key, value, merger);
                }
                else {
                    throw new Error(errors_1.default.IllegalStateException);
                }
            }
            else {
                map.put(key, value);
            }
        };
        var combiner = function (m1, m2) {
            m1.putAll(m2);
            return m1;
        };
        return Collector.of(supplier, accumulator, combiner, functions_1.Function.identity());
    };
    return Collectors;
}());
;
//return the largest of two values based on the comparator, first if they are equal
function returnLargest(first, second, comparator) {
    return comparator(first, second) < 0
        ? second
        : first;
}
//return the smallest of two values based on the comparator, first if they are equal
function returnSmallest(first, second, comparator) {
    return comparator(first, second) <= 0
        ? first
        : second;
}
exports.default = Collectors;
