"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var hash = require("js-hash-code");
var stream_1 = require("../stream");
var functions_1 = require("../functions");
var optional_1 = require("../optional");
var errors_1 = require("../errors");
exports.Map = {
    empty: function () {
        return new HashMap();
    },
    of: function (k1, v1, k2, v2, k3, v3, k4, v4, k5, v5) {
        var map = exports.Map.empty();
        var putIfBothPresent = function (k, v) {
            if (k !== undefined && v !== undefined) {
                map.put(k, v);
            }
        };
        putIfBothPresent(k1, v1);
        putIfBothPresent(k2, v2);
        putIfBothPresent(k3, v3);
        putIfBothPresent(k4, v4);
        putIfBothPresent(k5, v5);
        return map;
    }
};
exports.Entry = {
    of: function (key, value) {
        return new MapEntry(key, value);
    }
};
function throwIfNull(val) {
    optional_1.default.ofNullable(val)
        .orElseThrow(function () { return new Error(errors_1.Errors.NullPointerException); });
}
var HashMap = /** @class */ (function () {
    function HashMap() {
        this.map = {};
    }
    HashMap.prototype.clear = function () {
        this.map = {};
    };
    HashMap.prototype.isEmpty = function () {
        return this.keySet().length === 0;
    };
    HashMap.prototype.put = function (key, value) {
        throwIfNull(key);
        var keyHash = hash(key);
        var entry = exports.Entry.of(key, value);
        var previous = this.map[keyHash];
        this.map[keyHash] = entry;
        if (previous) {
            return previous.getValue();
        }
        else {
            return null;
        }
    };
    HashMap.prototype.putIfAbsent = function (key, value) {
        throwIfNull(key);
        var keyHash = hash(key);
        var entry = exports.Entry.of(key, value);
        var previous = this.map[keyHash];
        if (previous) {
            return previous.getValue();
        }
        else {
            this.map[keyHash] = entry;
            return null;
        }
    };
    HashMap.prototype.putAll = function (map) {
        var _this = this;
        map.forEach(function (key, value) { return _this.put(key, value); });
    };
    HashMap.prototype.get = function (key) {
        var keyHash = hash(key);
        var foundVal = this.map[keyHash];
        return foundVal ? foundVal.getValue() : null;
    };
    HashMap.prototype.getOptional = function (key) {
        var value = this.get(key);
        return value
            ? optional_1.default.of(value)
            : optional_1.default.empty();
    };
    HashMap.prototype.getOrDefault = function (key, defaultVal) {
        return this.getOptional(key)
            .orElse(defaultVal);
    };
    HashMap.prototype.values = function () {
        return this.entrySet()
            .map(function (e) { return e.getValue(); });
    };
    HashMap.prototype.keySet = function () {
        return this.entrySet()
            .map(function (e) { return e.getKey(); });
    };
    HashMap.prototype.entrySet = function () {
        var _this = this;
        return Object.keys(this.map)
            .map(function (key) { return _this.map[key]; });
    };
    HashMap.prototype.keyStream = function () {
        return stream_1.default.of(this.keySet());
    };
    HashMap.prototype.valueStream = function () {
        return stream_1.default.of(this.values());
    };
    HashMap.prototype.entryStream = function () {
        return stream_1.default.of(this.entrySet());
    };
    HashMap.prototype.containsKey = function (key) {
        return this.map.hasOwnProperty(hash(key));
    };
    HashMap.prototype.containsValue = function (value, equalityTest) {
        var equalityTestToUse = equalityTest ? equalityTest : functions_1.BiPredicate.defaultEquality();
        return this.valueStream()
            .anyMatch(function (v) { return equalityTestToUse(v, value); });
    };
    HashMap.prototype.forEach = function (consumer) {
        this.entrySet()
            .forEach(function (_a) {
            var key = _a.key, value = _a.value;
            return consumer(key, value);
        });
    };
    HashMap.prototype.remove = function (key) {
        var keyHash = hash(key);
        var previous = optional_1.default.ofNullable(this.map[keyHash]);
        delete this.map[keyHash];
        return previous
            .map(function (entry) { return entry.getValue(); })
            .orElseGet(function () { return null; });
    };
    HashMap.prototype.merge = function (key, value, remappingFunction) {
        var oldValue = this.getOptional(key);
        var newValue = oldValue.isPresent()
            ? remappingFunction(oldValue.get(), value)
            : value;
        if (newValue === null) {
            this.remove(key);
        }
        else {
            this.put(key, newValue);
        }
        return newValue;
    };
    return HashMap;
}());
var MapEntry = /** @class */ (function () {
    function MapEntry(key, value) {
        this.key = key;
        this.value = value;
    }
    MapEntry.prototype.getValue = function () {
        return this.value;
    };
    MapEntry.prototype.getKey = function () {
        return this.key;
    };
    return MapEntry;
}());
