"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var optional_1 = require("./optional");
exports.Processor = {
    mapProcessor: function (Function) { return new MapProcessor(Function); },
    filterProcessor: function (predicate) { return new FilterProcessor(predicate); },
    listFlatMapProcessor: function (Function) { return new ListFlatMapProcessor(Function); },
    distinctProcessor: function (comparator) { return new DistinctProcessor(comparator); },
    limitProcessor: function (limit) { return new LimitProcessor(limit); },
    streamFlatMapProcessor: function (Function) { return new StreamFlatMapProcessor(Function); },
    peekProcessor: function (consumer) { return new PeekProcessor(consumer); },
    optionalFlatMapProcessor: function (Function) { return new OptionalFlatMapProcessor(Function); },
    skipProcessor: function (numberToSkip) { return new SkipProcessor(numberToSkip); },
    sortProcessor: function (comparator) { return new SortProcessor(comparator); },
};
/**
 * Abstract processor that implements the storage and retrieval of items from
 * the processing queue.
 */
var AbstractProcessor = /** @class */ (function () {
    function AbstractProcessor() {
        this.inputs = [];
    }
    AbstractProcessor.prototype.add = function (input) {
        this.inputs.push(input);
    };
    AbstractProcessor.prototype.takeNextInput = function () {
        return optional_1.default.ofNullable(this.inputs.shift());
    };
    AbstractProcessor.prototype.hasNext = function () {
        return this.inputs.length > 0;
    };
    return AbstractProcessor;
}());
/**
 * Abstract processor that is stateless and not short circuiting
 */
var PureStatelessProcessor = /** @class */ (function (_super) {
    __extends(PureStatelessProcessor, _super);
    function PureStatelessProcessor() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    PureStatelessProcessor.prototype.isStateless = function () {
        return true;
    };
    PureStatelessProcessor.prototype.isShortCircuting = function () {
        return false;
    };
    return PureStatelessProcessor;
}(AbstractProcessor));
var LimitProcessor = /** @class */ (function (_super) {
    __extends(LimitProcessor, _super);
    function LimitProcessor(limit) {
        var _this = _super.call(this) || this;
        _this.limit = limit;
        _this.count = 0;
        return _this;
    }
    LimitProcessor.prototype.processAndGetNext = function () {
        if (this.count < this.limit) {
            this.count++;
            return this.takeNextInput();
        }
        else {
            return optional_1.default.empty();
        }
    };
    LimitProcessor.prototype.hasNext = function () {
        return (this.count < this.limit);
    };
    LimitProcessor.prototype.isStateless = function () {
        return true;
    };
    LimitProcessor.prototype.isShortCircuting = function () {
        return true;
    };
    return LimitProcessor;
}(AbstractProcessor));
/**
 * This is a stateful processor, that will return distinct elements provided all
 * the inputs are given at the start, and no elements are injected mid processing
 */
var DistinctProcessor = /** @class */ (function (_super) {
    __extends(DistinctProcessor, _super);
    function DistinctProcessor(comparator) {
        var _this = _super.call(this) || this;
        _this.comparator = comparator;
        _this.distinctList = [];
        return _this;
    }
    //distinct if stateful in the sense that it needs to keep track
    //of previous elements, but it does NOT need to greedily accumulate values
    DistinctProcessor.prototype.isStateless = function () {
        return true;
    };
    DistinctProcessor.prototype.isShortCircuting = function () {
        return false;
    };
    DistinctProcessor.prototype.hasNext = function () {
        return this.inputs.length > 0;
    };
    DistinctProcessor.prototype.processAndGetNext = function () {
        var _this = this;
        return this.takeNextInput().filter(function (i) { return _this.addIfUnique(i); });
    };
    /**
     * checks to see if the item is unique, adds it to the list if it is,
     * returns true if the value is distinct, otherwise false
     * @param item
     */
    DistinctProcessor.prototype.addIfUnique = function (item) {
        var _this = this;
        var doesMatchItem = function (distinct) { return _this.comparator(item, distinct); };
        var matchingItems = this.distinctList.filter(doesMatchItem);
        if (matchingItems.length === 0) {
            this.distinctList.push(item);
            return true;
        }
        else {
            return false;
        }
    };
    return DistinctProcessor;
}(AbstractProcessor));
var SortProcessor = /** @class */ (function (_super) {
    __extends(SortProcessor, _super);
    function SortProcessor(comparator) {
        var _this = _super.call(this) || this;
        _this.comparator = comparator;
        _this.sortedList = optional_1.default.empty();
        return _this;
    }
    SortProcessor.prototype.processValues = function () {
        var list = this.inputs.slice();
        list.sort(this.comparator);
        this.inputs = [];
        this.sortedList = optional_1.default.of(list);
    };
    SortProcessor.prototype.hasNext = function () {
        var sortedListExistsAndHasValues = this.sortedList.isPresent() ? this.sortedList.get().length > 0 : false;
        return this.inputs.length > 0 || sortedListExistsAndHasValues;
    };
    SortProcessor.prototype.processAndGetNext = function () {
        if (!this.sortedList.isPresent()) {
            this.processValues();
            return this.processAndGetNext();
        }
        else {
            return optional_1.default.ofNullable(this.sortedList.get().shift());
        }
    };
    SortProcessor.prototype.isStateless = function () {
        return false;
    };
    SortProcessor.prototype.isShortCircuting = function () {
        return false;
    };
    return SortProcessor;
}(AbstractProcessor));
/**
 * Implemention of a Processor for value mapping, lazily transforms values
 * when returned from the processor.
 */
var MapProcessor = /** @class */ (function (_super) {
    __extends(MapProcessor, _super);
    function MapProcessor(Function) {
        var _this = _super.call(this) || this;
        _this.Function = Function;
        return _this;
    }
    //pull values off the start
    MapProcessor.prototype.processAndGetNext = function () {
        return this.takeNextInput().map(this.Function);
    };
    return MapProcessor;
}(PureStatelessProcessor));
/**
 * Implemention of a Processor for consuming a value,intermediately but not not
 * altering the stream.
 */
var PeekProcessor = /** @class */ (function (_super) {
    __extends(PeekProcessor, _super);
    function PeekProcessor(consumer) {
        var _this = _super.call(this) || this;
        _this.consumer = consumer;
        return _this;
    }
    PeekProcessor.prototype.processAndGetNext = function () {
        var item = this.takeNextInput();
        item.ifPresent(this.consumer);
        return item;
    };
    return PeekProcessor;
}(PureStatelessProcessor));
/**
 * Stateless process, filters input items against a given predicated, only
 * returning those who match against the given predicate.
 */
var FilterProcessor = /** @class */ (function (_super) {
    __extends(FilterProcessor, _super);
    function FilterProcessor(predicate) {
        var _this = _super.call(this) || this;
        _this.predicate = predicate;
        return _this;
    }
    FilterProcessor.prototype.processAndGetNext = function () {
        return this.takeNextInput().filter(this.predicate);
    };
    return FilterProcessor;
}(PureStatelessProcessor));
/**
 * stateless processor, removes the given number of values before continuing to return values
 * given value must be positive, or nothing is skipped
 */
var SkipProcessor = /** @class */ (function (_super) {
    __extends(SkipProcessor, _super);
    function SkipProcessor(amountToSkip) {
        var _this = _super.call(this) || this;
        _this.amountToSkip = amountToSkip;
        return _this;
    }
    SkipProcessor.prototype.processAndGetNext = function () {
        if (this.amountToSkip > 0) {
            this.amountToSkip = this.amountToSkip - 1;
            this.takeNextInput(); //throw away
            return optional_1.default.empty();
        }
        else {
            return this.takeNextInput();
        }
    };
    return SkipProcessor;
}(PureStatelessProcessor));
/**
 * returns a one to many mapping of elements, transforms input elements into
 * a list of output elements, returning the values off the output lists, one list
 * at a time, lazily transforming inputs only when the previous list is exhausted.
 */
var ListFlatMapProcessor = /** @class */ (function (_super) {
    __extends(ListFlatMapProcessor, _super);
    function ListFlatMapProcessor(Function) {
        var _this = _super.call(this) || this;
        _this.Function = Function;
        _this.outputList = [];
        return _this;
    }
    ListFlatMapProcessor.prototype.hasNext = function () {
        return (this.outputList.length > 0 || this.inputs.length > 0);
    };
    ListFlatMapProcessor.prototype.processAndGetNext = function () {
        if (this.outputList.length > 0) {
            return optional_1.default.ofNullable(this.outputList.shift());
        }
        else if (this.inputs.length > 0) {
            var nextSource = this.takeNextInput();
            if (nextSource.isPresent()) {
                this.outputList = this.Function(nextSource.get());
                return this.processAndGetNext();
            }
        }
        return optional_1.default.empty();
    };
    return ListFlatMapProcessor;
}(PureStatelessProcessor));
/**
 * one to many mapping transformation, transforms elements through the stream bearing
 * transformation operation, and lazily returns elements off the resulting streams
 * one a time.
 */
var StreamFlatMapProcessor = /** @class */ (function (_super) {
    __extends(StreamFlatMapProcessor, _super);
    function StreamFlatMapProcessor(Function) {
        var _this = _super.call(this) || this;
        _this.Function = Function;
        return _this;
    }
    StreamFlatMapProcessor.prototype.hasNext = function () {
        return ((this.outputSpliterator && this.outputSpliterator.hasNext()) || this.inputs.length > 0);
    };
    StreamFlatMapProcessor.prototype.processAndGetNext = function () {
        if (this.outputSpliterator && this.outputSpliterator.hasNext()) {
            return this.outputSpliterator.getNext();
        }
        else if (this.inputs.length > 0) {
            var nextSource = this.takeNextInput();
            if (nextSource.isPresent()) {
                this.outputSpliterator = this.Function(nextSource.get()).streamIterator();
                return this.processAndGetNext();
            }
        }
        return optional_1.default.empty();
    };
    return StreamFlatMapProcessor;
}(PureStatelessProcessor));
/**
 * Processor that takes in source values, and transforms them through an optional bearing
 * Function, and returns the values in a flattened optional state
 */
var OptionalFlatMapProcessor = /** @class */ (function (_super) {
    __extends(OptionalFlatMapProcessor, _super);
    function OptionalFlatMapProcessor(Function) {
        var _this = _super.call(this) || this;
        _this.Function = Function;
        return _this;
    }
    OptionalFlatMapProcessor.prototype.processAndGetNext = function () {
        return this.takeNextInput().flatMap(this.Function);
    };
    return OptionalFlatMapProcessor;
}(PureStatelessProcessor));
