import { Function, Supplier } from "./functions";
import Optional from "./optional";
import Stream from "./stream";
/**
 * Defines the basic outline of a stream source, it needs to be iterable
 * and checkable as to if it has a next item.
 */
export interface Source<T> {
    /**
     * Returns a value if exists in the supplier, otherwise undefined. isEmpty() can be
     * used to reliably check if a value exists.
     */
    get(): T | undefined;
    /**
     * returns true if no more values will be returned by this supplier through get();
     */
    hasNext(): boolean;
}
export declare const Source: {
    /**
     * An infinite source that continually applies a function to a previous result, starting with the seed value
     *
     * seed, Function(seed), Function(Function(seed)), etc
     * @param seed initial starting value
     * @param Function function to be continually applied to the previous result, starting with the seed
     */
    iterateSource: <S>(seed: S, Function: Function<S, S>) => Source<S>;
    /**
    * basic infinite Source coming from a Supplier function
    */
    supplierSource: <S>(supplier: Supplier<S>) => Source<S>;
    /**
     * a source backed by the given array, it is not infinite
    */
    arraySource: <S>(array: S[]) => Source<S>;
    /**
     * source of values that first draws values from stream1 and then stream2
     */
    concatSource: <S>(stream1: Stream<S>, stream2: Stream<S>) => Source<Optional<S>>;
    /**
     * creates a source of numbers based on the given start and end bounds and the step size
     * a limited source counting by step (or 1) until the values get to the end bound.
     * steps are always considered as absolute values and step in the direction of the end bound. a step of -1 will
     * be treated the same as step of 1. if the start is larger than the end steps will be taken in a decreasing fashion
     * i.e. rangeSource(10, 5, -1) -> [10,9,8,7,6] as well as (10, 5, 1) will yield the same result
     *
     */
    rangeSource: (startInclusive: number, endExclusive: number, step?: number | undefined) => Source<number>;
};
