"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var chai_1 = require("chai");
var spies = require("chai-spies");
var stream_1 = require("../stream");
var collectors_1 = require("../collectors");
var __1 = require("..");
var functions_1 = require("../functions");
var errors_1 = require("../errors");
chai_1.use(spies);
describe('Collectors', function () {
    describe('toArray', function () {
        it('should collect the values to an array', function () {
            var source = [1, 3, 4];
            var stream = stream_1.default.of(source);
            var result = stream.collect(collectors_1.default.toArray());
            chai_1.expect(result).to.contain.ordered.members(source);
        });
    });
    describe('toList', function () {
        it('should collect the values to an array', function () {
            var source = [1, 3, 4];
            var stream = stream_1.default.of(source);
            var result = stream.collect(collectors_1.default.toList());
            chai_1.expect(result).to.contain.ordered.members(source);
        });
    });
    describe('joining', function () {
        it('should concatinate values into a single string', function () {
            var source = ["a", "b", "c"];
            var stream = stream_1.default.of(source);
            var result = stream.collect(collectors_1.default.joining());
            chai_1.expect(result).equal("abc");
        });
        it('should concatinate values into a single string, including a delimiter', function () {
            var source = ["a", "b", "c"];
            var stream = stream_1.default.of(source);
            var result = stream.collect(collectors_1.default.joining(","));
            chai_1.expect(result).equal("a,b,c");
        });
        it('should concatinate values into a single string, including delimiter, prefix and suffix', function () {
            var source = ["a", "b", "c"];
            var stream = stream_1.default.of(source);
            var result = stream.collect(collectors_1.default.joining(",", "pre-", "-post"));
            chai_1.expect(result).equal("pre-a,b,c-post");
        });
        it('should return empty string with no params and empty stream', function () {
            var stream = stream_1.default.empty();
            var result = stream.collect(collectors_1.default.joining());
            chai_1.expect(result).equal("");
        });
        it('should return empty string with delim params and empty stream', function () {
            var stream = stream_1.default.empty();
            var result = stream.collect(collectors_1.default.joining());
            chai_1.expect(result).equal("");
        });
        it('should return just pre and post when stream empty', function () {
            var stream = stream_1.default.empty();
            var result = stream.collect(collectors_1.default.joining(",", "a", "b"));
            chai_1.expect(result).equal("ab");
        });
    });
    describe('averagingNumber', function () {
        it('it should return the average of a given set of numbers', function () {
            var source = ["10", "20", "30"];
            var average = stream_1.default.of(source).collect(collectors_1.default.averagingNumber(parseInt));
            chai_1.expect(average).eq(20);
        });
        it('it should return the average of a given set of numbers', function () {
            var source = ["10", "11"];
            var average = stream_1.default.of(source).collect(collectors_1.default.averagingNumber(parseInt));
            chai_1.expect(average).eq(10.5);
        });
        it('it should return the average 0 if no inputs', function () {
            var source = [];
            var average = stream_1.default.of(source).collect(collectors_1.default.averagingNumber(parseInt));
            chai_1.expect(average).eq(0);
        });
    });
    describe('averaging', function () {
        it('it should return the average of a given set of numbers', function () {
            var source = ["10", "20", "30"];
            var average = stream_1.default.of(source)
                .map(parseInt)
                .collect(collectors_1.default.averaging());
            chai_1.expect(average).eq(20);
        });
    });
    describe('collectingAndThen', function () {
        it('it should return collector that applies new finisher as well', function () {
            var source = ["10", "20", "30"];
            var average = stream_1.default.of(source)
                .map(parseInt)
                .collect(collectors_1.default.collectingAndThen(collectors_1.default.averaging(), function (i) { return "" + i; }));
            chai_1.expect(average).eq("20");
        });
    });
    describe('counting', function () {
        it('it should return collector counts elements, given there is elements', function () {
            var source = ["10", "20", "30"];
            var count = stream_1.default.of(source)
                .collect(collectors_1.default.counting());
            chai_1.expect(count).eq(source.length);
        });
        it('it should return collector counts elements, given there is no elements', function () {
            var count = stream_1.default.empty()
                .collect(collectors_1.default.counting());
            chai_1.expect(count).eq(0);
        });
    });
    describe('groupingBy', function () {
        it('it should return collector that groups elements by classifier', function () {
            var source = ["cat", "dog", "ant", "dat", "car", "cut"];
            var mapByFirstLetter = stream_1.default.of(source)
                .collect(collectors_1.default.groupingBy(function (word) { return word.charAt(0); }));
            chai_1.expect(mapByFirstLetter.keySet().length).eq(3);
            chai_1.expect(mapByFirstLetter.getOptional('c').get()).to.contain.members(['cat', 'car', 'cut']);
            chai_1.expect(mapByFirstLetter.getOptional('d').get()).to.contain.members(['dog']);
            chai_1.expect(mapByFirstLetter.getOptional('a').get()).to.contain.members(['ant']);
        });
    });
    describe('mapping', function () {
        it('it should return collector that maps elements and then does the next collector', function () {
            var source = ["cat", "dog", "ant", "dat", "car", "cut"];
            var list = stream_1.default.of(source)
                .collect(collectors_1.default.mapping(function (s) { return s.length; }, collectors_1.default.toList()));
            chai_1.expect(list.length).eq(6);
            chai_1.expect(list).to.contain.members([3, 3, 3, 3, 3, 3]);
        });
    });
    describe('maxBy', function () {
        it('it should return collector that returns largest element by comparator, default comparator', function () {
            var source = [6, 12, 5, 1, 3];
            var max = stream_1.default.of(source)
                .collect(collectors_1.default.maxBy());
            chai_1.expect(max.isPresent()).eq(true);
            chai_1.expect(max.get()).to.be.eq(12);
        });
        it('it should return collector that returns largest element by comparator, custom comparator', function () {
            var reverseComparator = function (x, y) { return __1.Comparator.default()(y, x); };
            var source = [6, 12, 5, 1, 3];
            var max = stream_1.default.of(source)
                .collect(collectors_1.default.maxBy(reverseComparator));
            chai_1.expect(max.isPresent()).eq(true);
            chai_1.expect(max.get()).to.be.eq(1);
        });
        it('it should return collector result is empty optional if no elements', function () {
            var max = stream_1.default.empty()
                .collect(collectors_1.default.maxBy());
            chai_1.expect(max.isPresent()).eq(false);
        });
    });
    describe('minBy', function () {
        it('it should return collector that returns smallest element by comparator, default comparator', function () {
            var source = [6, 12, 5, 1, 3];
            var min = stream_1.default.of(source)
                .collect(collectors_1.default.minBy());
            chai_1.expect(min.isPresent()).eq(true);
            chai_1.expect(min.get()).to.be.eq(1);
        });
        it('it should return collector that returns largest element by comparator, custom comparator', function () {
            var reverseComparator = function (x, y) { return __1.Comparator.default()(y, x); };
            var source = [6, 12, 5, 1, 3];
            var min = stream_1.default.of(source)
                .collect(collectors_1.default.minBy(reverseComparator));
            chai_1.expect(min.isPresent()).eq(true);
            chai_1.expect(min.get()).to.be.eq(12);
        });
        it('it should return collector result is empty optional if no elements', function () {
            var min = stream_1.default.empty()
                .collect(collectors_1.default.minBy());
            chai_1.expect(min.isPresent()).eq(false);
        });
    });
    describe('partitioningBy', function () {
        it('it should return collector partitions by predicate without downstream', function () {
            var source = ["cat", "dogs", "ants", "dat", "car", "cut"];
            var map = stream_1.default.of(source)
                .collect(collectors_1.default.partitioningBy(function (i) { return i.length === 4; }));
            chai_1.expect(map.containsKey(true)).eq(true);
            chai_1.expect(map.containsKey(false)).eq(true);
            chai_1.expect(map.get(true)).to.include.members(["dogs", "ants"]);
            chai_1.expect(map.get(false)).to.include.members(["cat", "dat", "car", "cut"]);
        });
        it('it should return empty lists for when nothing in partition', function () {
            var source = ["cat", "dat", "car", "cut"];
            var map = stream_1.default.of(source)
                .collect(collectors_1.default.partitioningBy(function (i) { return i.length === 4; }));
            chai_1.expect(map.containsKey(true)).eq(true);
            chai_1.expect(map.containsKey(false)).eq(true);
            chai_1.expect(map.getOptional(true).get().length).to.eq(0);
            chai_1.expect(map.get(false)).to.include.members(["cat", "dat", "car", "cut"]);
        });
        it('it should return collector partitions by predicate with downstream', function () {
            var source = ["cat", "dogs", "ants", "dat", "car", "cut"];
            var map = stream_1.default.of(source)
                .collect(collectors_1.default.partitioningBy(function (i) { return i.length === 4; }, collectors_1.default.joining("")));
            chai_1.expect(map.containsKey(true)).eq(true);
            chai_1.expect(map.containsKey(false)).eq(true);
            chai_1.expect(map.get(true)).to.eq("dogsants");
            chai_1.expect(map.get(false)).to.eq("catdatcarcut");
        });
        it('it should return collector partitions by predicate with downstream, if empty returns collector default', function () {
            var source = ["cat", "dat", "car", "cut"];
            var map = stream_1.default.of(source)
                .map(function (i) { return i.length; })
                .collect(collectors_1.default.partitioningBy(function (i) { return i === 4; }, collectors_1.default.averaging()));
            chai_1.expect(map.containsKey(true)).eq(true);
            chai_1.expect(map.containsKey(false)).eq(true);
            chai_1.expect(map.get(true)).to.eq(0);
            chai_1.expect(map.get(false)).to.eq(3);
        });
    });
    describe('reducing', function () {
        it('it should work if just given a reducing function', function () {
            var source = [1, 2, 3, 4];
            var result = stream_1.default.of(source)
                .collect(collectors_1.default.reducing(function (i, i2) { return i + i2; }));
            chai_1.expect(result.isPresent()).eq(true);
            chai_1.expect(result.get()).eq(10);
        });
        it('it should work if just given a reducing function, empty inputs', function () {
            var result = stream_1.default.empty()
                .collect(collectors_1.default.reducing(function (i, i2) { return i + i2; }));
            chai_1.expect(result.isPresent()).eq(false);
        });
        it('it should work if just given a reducing function and identity, empty inputs return identity', function () {
            var id = 5;
            var result = stream_1.default.empty()
                .collect(collectors_1.default.reducing(function (i, i2) { return i + i2; }, id));
            chai_1.expect(result.isPresent()).eq(true);
            chai_1.expect(result.get()).eq(id);
        });
        it('it should work if just given a reducing function and identty includes identity', function () {
            var id = 1;
            var source = [1, 2, 3, 4];
            var result = stream_1.default.of(source)
                .collect(collectors_1.default.reducing(function (i, i2) { return i + i2; }, id));
            chai_1.expect(result.isPresent()).eq(true);
            chai_1.expect(result.get()).eq(10 + id);
        });
        it('it should work if just given a reducing function and identty includes identity', function () {
            var id = 1;
            var source = ['1', '2', '3', '4'];
            var result = stream_1.default.of(source)
                .collect(collectors_1.default.reducing(function (i, i2) { return i + i2; }, id, function (s) { return parseInt(s); }));
            chai_1.expect(result.isPresent()).eq(true);
            chai_1.expect(result.get()).eq(10 + id);
        });
    });
    describe('summarizingNumber', function () {
        it('it should return collector applies to numbers if no mapper specified', function () {
            var source = ["10", "20", "30"];
            var average = stream_1.default.of(source)
                .map(parseInt)
                .collect(collectors_1.default.summarizingNumber());
            chai_1.expect(average.getAverage()).eq(20);
            chai_1.expect(average.getMax()).eq(30);
            chai_1.expect(average.getMin()).eq(10);
            chai_1.expect(average.getSum()).eq(60);
            chai_1.expect(average.getCount()).eq(3);
        });
        it('it should return collector applies to non numbers if mapper specified', function () {
            var source = ["10", "20", "30"];
            var average = stream_1.default.of(source)
                .collect(collectors_1.default.summarizingNumber(parseInt));
            chai_1.expect(average.getAverage()).eq(20);
            chai_1.expect(average.getMax()).eq(30);
            chai_1.expect(average.getMin()).eq(10);
            chai_1.expect(average.getSum()).eq(60);
            chai_1.expect(average.getCount()).eq(3);
        });
        it('it should return collector applies to numbers if mapper specified', function () {
            var source = [10, 20, 30];
            var average = stream_1.default.of(source)
                .collect(collectors_1.default.summarizingNumber(functions_1.Function.identity()));
            chai_1.expect(average.getAverage()).eq(20);
            chai_1.expect(average.getMax()).eq(30);
            chai_1.expect(average.getMin()).eq(10);
            chai_1.expect(average.getSum()).eq(60);
            chai_1.expect(average.getCount()).eq(3);
        });
    });
    describe('summingNumber', function () {
        it('it should return collector that sums numbers without a mapper', function () {
            var source = ["10", "20", "30"];
            var sum = stream_1.default.of(source)
                .map(parseInt)
                .collect(collectors_1.default.summingNumber());
            chai_1.expect(sum).eq(60);
        });
        it('it should return collector that sums with a mapper ', function () {
            var source = ["10", "20", "30"];
            var sum = stream_1.default.of(source)
                .collect(collectors_1.default.summingNumber(parseInt));
            chai_1.expect(sum).eq(60);
        });
        it('it should return collector applies to numbers with mapper and number stream', function () {
            var source = ["10", "20", "30"];
            var sum = stream_1.default.of(source)
                .map(parseInt)
                .collect(collectors_1.default.summingNumber(function (i) { return i + 1; }));
            chai_1.expect(sum).eq(63);
        });
    });
    describe('toMap', function () {
        it('saves all items with the mapped keys and values', function () {
            var source = ["1", "2", "3"];
            var stream = stream_1.default.of(source);
            var result = stream.collect(collectors_1.default.toMap(function (str) { return "key:" + str; }, parseInt));
            chai_1.expect(result.get("key:1")).to.eq(1);
            chai_1.expect(result.get("key:2")).to.eq(2);
            chai_1.expect(result.get("key:3")).to.eq(3);
        });
        it('saves all items throws error without mapper and dupe keys', function () {
            var source = ["1", "2", "2"];
            var stream = stream_1.default.of(source);
            var action = function () { return stream.collect(collectors_1.default.toMap(function (str) { return "key:" + str; }, parseInt)); };
            chai_1.expect(action).to.throw(errors_1.default.IllegalStateException);
        });
        it('saves all items and merges with merger when there is dupes', function () {
            var source = ["1", "2", "2"];
            var stream = stream_1.default.of(source);
            var resultMap = stream.collect(collectors_1.default.toMap(function (str) { return "key:" + str; }, parseInt, function (i, i2) { return i + i2; }));
            chai_1.expect(resultMap.get("key:1")).to.eq(1);
            chai_1.expect(resultMap.get("key:2")).to.eq(4);
        });
    });
});
