"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var map_1 = require("../map");
var errors_1 = require("../errors");
var functions_1 = require("../functions");
var chai_1 = require("chai");
var spies = require("chai-spies");
chai_1.use(spies);
var testObj = function (v1, v2) {
    return { a: v1, b: { x: v2 } };
};
describe('Map tests', function () {
    describe('of', function () {
        it('it creates a map with specified key value pairs', function () {
            var map = map_1.Map.of('one', 1, 'two', 2, 'three', 3, 'four', 4, 'five', 5);
            chai_1.expect(map.keySet().length).eq(5);
            chai_1.expect(map.get('one')).to.be.eq(1);
            chai_1.expect(map.get('two')).to.be.eq(2);
            chai_1.expect(map.get('three')).to.be.eq(3);
            chai_1.expect(map.get('four')).to.be.eq(4);
            chai_1.expect(map.get('five')).to.be.eq(5);
        });
        it('it can create empty', function () {
            var map = map_1.Map.of();
            chai_1.expect(map.isEmpty()).to.be.true;
        });
        it('it doesnt add malformed pair', function () {
            var map = map_1.Map.of('one', 1, 'two');
            chai_1.expect(map.keySet().length).eq(1);
            chai_1.expect(map.getOptional('two').isPresent()).to.be.false;
        });
    });
    describe('empty', function () {
        it('it creates empty map', function () {
            var map = map_1.Map.empty();
            chai_1.expect(map.isEmpty()).to.be.true;
        });
    });
    describe('clear', function () {
        it('it removes all keys and values from the map', function () {
            var map = map_1.Map.of('one', 1, 'two', 2, 'three', 3, 'four', 4, 'five', 5);
            chai_1.expect(map.isEmpty()).to.be.false;
            map.clear();
            chai_1.expect(map.isEmpty()).to.be.true;
        });
    });
    describe('get', function () {
        it('it returns the value in the map for the key', function () {
            var key = 'one';
            var value = 100;
            var map = map_1.Map.of(key, value);
            chai_1.expect(map.get(key)).to.eq(value);
        });
        it('it returns the null for unmatched key', function () {
            var key = 'one';
            var value = 100;
            var map = map_1.Map.of(key, value);
            chai_1.expect(map.get('notkey')).to.eq(null);
        });
    });
    describe('getOptional', function () {
        it('it returns optional describing the value for a given key', function () {
            var key = 'one';
            var value = 100;
            var map = map_1.Map.of(key, value);
            chai_1.expect(map.getOptional(key).isPresent()).to.be.true;
            chai_1.expect(map.getOptional(key).get()).to.be.eq(value);
        });
        it('it returns optional describing the value for a given key', function () {
            var key = 'one';
            var value = 100;
            var map = map_1.Map.of(key, value);
            chai_1.expect(map.getOptional('notkey').isPresent()).to.be.false;
        });
    });
    describe('getOrDefault', function () {
        it('it returns value if one exists', function () {
            var key = 'one';
            var value = 100;
            var def = 200;
            var map = map_1.Map.of(key, value);
            chai_1.expect(map.getOrDefault(key, def)).to.be.eq(value);
        });
        it('it returns default if value was not there', function () {
            var key = 'one';
            var value = 100;
            var def = 200;
            var map = map_1.Map.of(key, value);
            chai_1.expect(map.getOrDefault('notkey', def)).to.be.eq(def);
        });
    });
    describe('put', function () {
        it('add entry to map with basic types', function () {
            var key = "key";
            var val = "val";
            var map = map_1.Map.empty();
            map.put(key, val);
            var getResult = map.get(key);
            chai_1.expect(getResult).to.be.equal(val);
        });
        it('throw error if key is null', function () {
            var key = null;
            var val = "val";
            var map = map_1.Map.empty();
            chai_1.expect(function () { return map.put(key, val); }).to.throw(errors_1.Errors.NullPointerException);
        });
        it('throw error if key is undefined', function () {
            var key = undefined;
            var val = "val";
            var map = map_1.Map.empty();
            chai_1.expect(function () { return map.put(key, val); }).to.throw(errors_1.Errors.NullPointerException);
        });
        it('add replace entry with basic types', function () {
            var key = "key";
            var val = "val";
            var val2 = "val2";
            var map = map_1.Map.empty();
            map.put(key, val);
            var result = map.put(key, val2);
            var getResult = map.get(key);
            chai_1.expect(result).to.equal(val);
            chai_1.expect(getResult).to.be.equal(val2);
        });
        it('add value with complex type key', function () {
            var key = testObj("test1", 5);
            var val = testObj("val1", 6);
            var map = map_1.Map.empty();
            map.put(key, val);
            var getResult = map.get(key);
            chai_1.expect(getResult).to.be.equal(val);
        });
        it('add object with same fields and values should create same hash', function () {
            var key = testObj("test1", 5);
            var key2 = testObj("test1", 5);
            var val = testObj("val1", 6);
            var val2 = testObj("val2", 7);
            var map = map_1.Map.empty();
            map.put(key, val);
            map.put(key2, val2);
            var getResult = map.get(key);
            chai_1.expect(getResult).to.be.equal(val2);
        });
    });
    describe('putIfAbsent', function () {
        it('throw error if key is null', function () {
            var key = null;
            var val = "val";
            var map = map_1.Map.empty();
            chai_1.expect(function () { return map.putIfAbsent(key, val); }).to.throw(errors_1.Errors.NullPointerException);
        });
        it('throw error if key is undefined', function () {
            var key = undefined;
            var val = "val";
            var map = map_1.Map.empty();
            chai_1.expect(function () { return map.putIfAbsent(key, val); }).to.throw(errors_1.Errors.NullPointerException);
        });
        it('it adds a new mapping if the specified key is not already in use', function () {
            var key = "key";
            var val = "val";
            var key2 = "key2";
            var val2 = "val2";
            var map = map_1.Map.of(key, val);
            map.putIfAbsent(key2, val2);
            chai_1.expect(map.keySet().length).to.be.eq(2);
            chai_1.expect(map.get(key2)).to.be.equal(val2);
            chai_1.expect(map.get(key)).to.be.equal(val);
        });
        it('it returns null if no current mapping', function () {
            var key = "key";
            var val = "val";
            var key2 = "key2";
            var val2 = "val2";
            var map = map_1.Map.of(key, val);
            var retVal = map.putIfAbsent(key2, val2);
            chai_1.expect(retVal).to.be.equal(null);
        });
        it('it does nothing if key already present', function () {
            var key = "key";
            var val = "val";
            var val2 = "val2";
            var map = map_1.Map.of(key, val);
            map.putIfAbsent(key, val2);
            chai_1.expect(map.keySet().length).to.be.eq(1);
            chai_1.expect(map.get(key)).to.be.equal(val);
        });
        it('it returns the current mapping value if key was already present', function () {
            var key = "key";
            var val = "val";
            var val2 = "val2";
            var map = map_1.Map.of(key, val);
            var retVal = map.putIfAbsent(key, val2);
            chai_1.expect(retVal).to.be.equal(val);
        });
    });
    describe('putAll', function () {
        it('adds all the items from the given map to the map', function () {
            var map = map_1.Map.of('one', 1, 'two', 2);
            var map2 = map_1.Map.of('three', 3, 'four', 4);
            map.putAll(map2);
            chai_1.expect(map.valueStream().count()).to.eq(4);
            chai_1.expect(map.get('three')).to.eq(3);
            chai_1.expect(map.get('four')).to.eq(4);
        });
    });
    describe('containsKey', function () {
        it('returns true if the map contains the given key', function () {
            var key = 420;
            var map = map_1.Map.of(key, 'val');
            chai_1.expect(map.containsKey(key)).to.be.true;
        });
        it('returns false if the map not contains the given key', function () {
            var key = 420;
            var map = map_1.Map.of(key, 'val');
            chai_1.expect(map.containsKey(69)).to.be.false;
        });
        it('returns true if the map contains the given key, complex object', function () {
            var key = new Date(1);
            var secondKey = new Date(1);
            var map = map_1.Map.of(key, 'val');
            chai_1.expect(map.containsKey(secondKey)).to.be.true;
        });
    });
    describe('merge', function () {
        it('it adds a new value when key doesnt exist', function () {
            var key = "one";
            var value = 1;
            var map = map_1.Map.empty();
            map.merge(key, value, function (n1, n2) { return n1 + n2; });
            var result = map.getOptional(key);
            chai_1.expect(result.isPresent()).to.be.true;
            chai_1.expect(result.get()).to.be.eq(value);
        });
        it('it adds a new value when key doesnt exist', function () {
            var key = "one";
            var value = 1;
            var map = map_1.Map.empty();
            map.put(key, value);
            map.merge(key, 1, function (n1, n2) { return n1 + n2; });
            var result = map.getOptional(key);
            chai_1.expect(result.isPresent()).to.be.true;
            chai_1.expect(result.get()).to.be.eq(value + value);
        });
        it('it removes entry when remapper returns null', function () {
            var key = "one";
            var value = 1;
            var map = map_1.Map.empty();
            map.put(key, value);
            map.merge(key, 1, function (n1, n2) { return null; });
            var result = map.getOptional(key);
            chai_1.expect(result.isPresent()).to.be.false;
            chai_1.expect(map.containsKey(key)).to.be.false;
        });
    });
    describe('containsKey', function () {
        it('it returns true when a key exists', function () {
            var key = "key";
            var map = map_1.Map.of(key, 'val');
            chai_1.expect(map.containsKey(key)).to.be.true;
        });
        it('it returns true when a key exists', function () {
            var key = new Date();
            var map = map_1.Map.of(key, 'val');
            chai_1.expect(map.containsKey(key)).to.be.true;
        });
        it('it returns false when a key doesnt exists', function () {
            var key = new Date(1);
            var keyOther = new Date(100000000);
            var map = map_1.Map.of(key, 'val');
            chai_1.expect(map.containsKey(keyOther)).to.be.false;
        });
        it('it returns false when a key given is null', function () {
            var key = new Date(1);
            var keyOther = null;
            var map = map_1.Map.of(key, 'val');
            chai_1.expect(map.containsKey(keyOther)).to.be.false;
        });
    });
    describe('containsValue', function () {
        it('it returns true if value exists', function () {
            var value = "100";
            var map = map_1.Map.of(1, value, 2, value);
            chai_1.expect(map.containsValue(value)).to.be.true;
        });
        it('it returns false if value doesnt xists', function () {
            var value = "100";
            var map = map_1.Map.of(1, value, 2, value);
            chai_1.expect(map.containsValue('notValue')).to.be.false;
        });
        it('it returns true if value exists based on equality function', function () {
            var value = new Date(1000);
            var value2 = new Date(1000);
            var map = map_1.Map.of(1, value, 2, value);
            chai_1.expect(map.containsValue(value2)).to.be.false; //fail default equality
            chai_1.expect(map.containsValue(value2, functions_1.BiPredicate.hashEquality())).to.be.true;
        });
        it('it returns false if value doesnt exists based on equality function', function () {
            var value = new Date(1000);
            var value2 = new Date(200000);
            var map = map_1.Map.of(1, value, 2, value);
            chai_1.expect(map.containsValue(value2, functions_1.BiPredicate.hashEquality())).to.be.false;
        });
    });
    describe('keySet', function () {
        it('returns an array of all the keys in the map', function () {
            var keys = [1, 2, 3];
            var map = map_1.Map.of(keys[0], "1", keys[1], "2", keys[2], "3");
            var result = map.keySet();
            chai_1.expect(result).to.have.members(keys);
        });
        it('returns an empty array if map is empty', function () {
            var map = map_1.Map.empty();
            var result = map.keySet();
            chai_1.expect(result.length).to.be.eq(0);
        });
    });
    describe('values', function () {
        it('returns an array of all the values in the map', function () {
            var values = [1, 2, 3];
            var map = map_1.Map.of(1, values[0], 2, values[1], 3, values[2]);
            var result = map.values();
            chai_1.expect(result).to.have.members(values);
        });
        it('returns an empty array if map is empty', function () {
            var map = map_1.Map.empty();
            var result = map.values();
            chai_1.expect(result.length).to.be.eq(0);
        });
    });
    describe('keyStream', function () {
        it('returns an array of all the keys in the map', function () {
            var keys = [1, 2, 3];
            var map = map_1.Map.of(keys[0], "1", keys[1], "2", keys[2], "3");
            var result = map.keyStream().toArray();
            chai_1.expect(result).to.have.members(keys);
        });
        it('returns an empty array if map is empty', function () {
            var map = map_1.Map.empty();
            var result = map.keyStream().toArray();
            chai_1.expect(result.length).to.be.eq(0);
        });
    });
    describe('valueStream', function () {
        it('returns an array of all the values in the map', function () {
            var values = [1, 2, 3];
            var map = map_1.Map.of(1, values[0], 2, values[1], 3, values[2]);
            var result = map.valueStream().toArray();
            chai_1.expect(result).to.have.members(values);
        });
        it('returns an empty array if map is empty', function () {
            var map = map_1.Map.empty();
            var result = map.valueStream().toArray();
            chai_1.expect(result.length).to.be.eq(0);
        });
    });
    describe('entrySet', function () {
        it('returns an array of all the entrys in the map', function () {
            var values = [1, 2, 3];
            var keys = [4, 5, 6];
            var map = map_1.Map.of(keys[0], values[0], keys[1], values[1], keys[2], values[2]);
            var result = map.entrySet();
            chai_1.expect(result.length).to.eq(3);
            chai_1.expect(result[0].key).to.be.eq(keys[0]);
            chai_1.expect(result[0].value).to.be.eq(values[0]);
            chai_1.expect(result[1].key).to.be.eq(keys[1]);
            chai_1.expect(result[1].value).to.be.eq(values[1]);
            chai_1.expect(result[2].key).to.be.eq(keys[2]);
            chai_1.expect(result[2].value).to.be.eq(values[2]);
        });
        it('returns an empty array if map is empty', function () {
            var map = map_1.Map.empty();
            var result = map.entrySet();
            chai_1.expect(result.length).to.be.eq(0);
        });
    });
    describe('entryStream', function () {
        it('returns an stream of all the entrys in the map', function () {
            var values = [1, 2, 3];
            var keys = [4, 5, 6];
            var map = map_1.Map.of(keys[0], values[0], keys[1], values[1], keys[2], values[2]);
            var result = map.entryStream().toArray();
            chai_1.expect(result.length).to.eq(3);
            chai_1.expect(result[0].key).to.be.eq(keys[0]);
            chai_1.expect(result[0].value).to.be.eq(values[0]);
            chai_1.expect(result[1].key).to.be.eq(keys[1]);
            chai_1.expect(result[1].value).to.be.eq(values[1]);
            chai_1.expect(result[2].key).to.be.eq(keys[2]);
            chai_1.expect(result[2].value).to.be.eq(values[2]);
        });
        it('returns an empty array if map is empty', function () {
            var map = map_1.Map.empty();
            var result = map.entryStream().toArray();
            chai_1.expect(result.length).to.be.eq(0);
        });
    });
    describe('forEach', function () {
        it('calls function for each entry in the map', function () {
            var map = map_1.Map.of(1, 1, 2, 2, 3, 3);
            var consumer = function (x, y) { return x + y; };
            var spyConsumer = chai_1.spy(consumer);
            map.forEach(spyConsumer);
            chai_1.expect(spyConsumer).to.be.called.with(1, 1);
            chai_1.expect(spyConsumer).to.be.called.with(2, 2);
            chai_1.expect(spyConsumer).to.be.called.with(3, 3);
        });
    });
    describe('isEmpty', function () {
        it('returns true when empty', function () {
            var map = map_1.Map.empty();
            chai_1.expect(map.isEmpty()).to.be.true;
        });
        it('returns true when empty by removal', function () {
            var map = map_1.Map.of(1, 1);
            map.remove(1);
            chai_1.expect(map.isEmpty()).to.be.true;
        });
        it('returns false when has values', function () {
            var map = map_1.Map.of(1, 1);
            chai_1.expect(map.isEmpty()).to.be.false;
        });
    });
    describe('remove', function () {
        it('removes the given key from the map, if it existed', function () {
            var key = 1;
            var map = map_1.Map.of(key, 1);
            map.remove(key);
            chai_1.expect(map.containsKey(key)).to.be.false;
        });
        it('returns previous value if one existed', function () {
            var key = 1;
            var val = '1';
            var map = map_1.Map.of(key, val);
            var result = map.remove(key);
            chai_1.expect(result).to.be.eq(val);
        });
        it('returns null value if one no key existed', function () {
            var key = 1;
            var val = '1';
            var map = map_1.Map.of(key, val);
            var result = map.remove(2);
            chai_1.expect(result).to.be.null;
        });
    });
});
