"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var chai_1 = require("chai");
var spies = require("chai-spies");
var functions_1 = require("../functions");
var processor_1 = require("../processor");
var processorPipeline_1 = require("../processorPipeline");
chai_1.use(spies);
describe('ProcessorPipeline tests', function () {
    describe('ProcessorNode tests', function () {
        it('should create new node with empty previous and next', function () {
            var processor = processor_1.Processor.mapProcessor(functions_1.Function.identity());
            var processorNode = new processorPipeline_1.ProcessorNode(processor);
            chai_1.expect(processorNode.getNextNode().isPresent()).is.equal(false);
            chai_1.expect(processorNode.getPreviousNode().isPresent()).is.equal(false);
        });
        it('addNext should add a processor to next', function () {
            var processor = processor_1.Processor.mapProcessor(functions_1.Function.identity());
            var processorNode = new processorPipeline_1.ProcessorNode(processor);
            var nextProcessor = new processorPipeline_1.ProcessorNode(processor);
            processorNode.addNextNode(nextProcessor);
            chai_1.expect(processorNode.getNextNode().isPresent()).is.equal(true);
            chai_1.expect(processorNode.getNextNode().get()).is.equal(nextProcessor);
        });
        it('addPrev should add a processor to next', function () {
            var processor = processor_1.Processor.mapProcessor(functions_1.Function.identity());
            var processorNode = new processorPipeline_1.ProcessorNode(processor);
            var prev = new processorPipeline_1.ProcessorNode(processor);
            processorNode.addPreviousNode(prev);
            chai_1.expect(processorNode.getPreviousNode().isPresent()).is.equal(true);
            chai_1.expect(processorNode.getPreviousNode().get()).is.equal(prev);
        });
        it('has next false when short circuiting, processor does not have next, but there is previous inputs', function () {
            var processor = processor_1.Processor.limitProcessor(0);
            var processorNode = new processorPipeline_1.ProcessorNode(processor);
            var prev = new processorPipeline_1.ProcessorNode(processor_1.Processor.mapProcessor(functions_1.Function.identity()));
            prev.addInput('1');
            processorNode.addPreviousNode(prev);
            chai_1.expect(processorNode.hasNext()).is.equal(false);
        });
        it('has next false when short circuiting, processor does not have next, and no inputs from prev', function () {
            var processor = processor_1.Processor.limitProcessor(0);
            var processorNode = new processorPipeline_1.ProcessorNode(processor);
            var prev = new processorPipeline_1.ProcessorNode(processor_1.Processor.mapProcessor(functions_1.Function.identity()));
            processorNode.addPreviousNode(prev);
            chai_1.expect(processorNode.hasNext()).is.equal(false);
        });
        it('has next true when short circuiting, processor has have next, and has inputs from prev', function () {
            var processor = processor_1.Processor.limitProcessor(1);
            var processorNode = new processorPipeline_1.ProcessorNode(processor);
            var prev = new processorPipeline_1.ProcessorNode(processor_1.Processor.mapProcessor(functions_1.Function.identity()));
            prev.addInput("1");
            processorNode.addPreviousNode(prev);
            chai_1.expect(processorNode.hasNext()).is.equal(true);
        });
        it('has next false when short circuiting, processor has have next but previous node is empty', function () {
            var processor = processor_1.Processor.limitProcessor(1);
            var processorNode = new processorPipeline_1.ProcessorNode(processor);
            var prev = new processorPipeline_1.ProcessorNode(processor_1.Processor.mapProcessor(functions_1.Function.identity()));
            processorNode.addPreviousNode(prev);
            chai_1.expect(processorNode.hasNext()).is.equal(false);
        });
        it('statefulPullAndGet fills processor from previous if previous exists', function () {
            var processor = processor_1.Processor.sortProcessor(functions_1.Comparator.default());
            var processorNode = new processorPipeline_1.ProcessorNode(processor);
            var prev = new processorPipeline_1.ProcessorNode(processor_1.Processor.mapProcessor(functions_1.Function.identity()));
            prev.addInput("a");
            prev.addInput("b");
            processorNode.addPreviousNode(prev);
            chai_1.expect(processorNode.statefulPullAndGet().get()).is.equal("a");
            chai_1.expect(processorNode.statefulPullAndGet().get()).is.equal("b");
        });
        it('statefulPullAndGet gets from processors values if no previous', function () {
            var processor = processor_1.Processor.sortProcessor(functions_1.Comparator.default());
            var processorNode = new processorPipeline_1.ProcessorNode(processor);
            processorNode.addInput("a");
            chai_1.expect(processorNode.statefulPullAndGet().get()).is.equal("a");
        });
        it('statelessGet pulls first from current Processor if it has values', function () {
            var processor = processor_1.Processor.mapProcessor(functions_1.Function.identity());
            var processorNode = new processorPipeline_1.ProcessorNode(processor);
            processorNode.addInput("a");
            var prev = new processorPipeline_1.ProcessorNode(processor_1.Processor.mapProcessor(functions_1.Function.identity()));
            prev.addInput("b");
            processorNode.addPreviousNode(prev);
            chai_1.expect(processorNode.statelessGet().get()).is.equal("a");
        });
        it('statelessGet pulls from previous processor if current processor is empty', function () {
            var processor = processor_1.Processor.mapProcessor(functions_1.Function.identity());
            var processorNode = new processorPipeline_1.ProcessorNode(processor);
            var prev = new processorPipeline_1.ProcessorNode(processor_1.Processor.mapProcessor(functions_1.Function.identity()));
            prev.addInput("b");
            processorNode.addPreviousNode(prev);
            chai_1.expect(processorNode.statelessGet().get()).is.equal("b");
        });
        it('statelessGet returns Optional.empty if previous and current processors are empty', function () {
            var processor = processor_1.Processor.mapProcessor(functions_1.Function.identity());
            var processorNode = new processorPipeline_1.ProcessorNode(processor);
            var prev = new processorPipeline_1.ProcessorNode(processor_1.Processor.mapProcessor(functions_1.Function.identity()));
            processorNode.addPreviousNode(prev);
            chai_1.expect(processorNode.statelessGet().isPresent()).is.equal(false);
        });
        it('getProcessedValue calls statelessGet when node is stateless', function () {
            var processor = processor_1.Processor.mapProcessor(functions_1.Function.identity());
            var processorNode = new processorPipeline_1.ProcessorNode(processor);
            var statelessGetSpy = chai_1.spy.on(processorNode, 'statelessGet');
            processorNode.getProcessedValue();
            chai_1.expect(statelessGetSpy).to.have.been.called();
        });
        it('getProcessedValue calls statefulPullAndGet when node is stateful', function () {
            var processor = processor_1.Processor.sortProcessor(functions_1.Comparator.default());
            var processorNode = new processorPipeline_1.ProcessorNode(processor);
            var statelessGetSpy = chai_1.spy.on(processorNode, 'statefulPullAndGet');
            processorNode.getProcessedValue();
            chai_1.expect(statelessGetSpy).to.have.been.called();
        });
    });
});
