"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var functions_1 = require("../functions");
var optional_1 = require("../optional");
var processor_1 = require("../processor");
var chai_1 = require("chai");
var stream_1 = require("../stream");
describe('Processor tests', function () {
    describe('MapProcessor tests', function () {
        it('should be a stateless processor', function () {
            var Function = function (n) { return "" + n; };
            var processor = processor_1.Processor.mapProcessor(Function);
            chai_1.expect(processor.isStateless()).to.equal(true);
        });
        it('should not have next with no values', function () {
            var Function = function (n) { return "" + n; };
            var processor = processor_1.Processor.mapProcessor(Function);
            chai_1.expect(processor.hasNext()).to.equal(false);
        });
        it('should have next when a value as been added', function () {
            var Function = function (n) { return "" + n; };
            var processor = processor_1.Processor.mapProcessor(Function);
            processor.add(1);
            chai_1.expect(processor.hasNext()).to.equal(true);
        });
        it('should run values through Function when retrieved', function () {
            var Function = function (n) { return "" + n; };
            var processor = processor_1.Processor.mapProcessor(Function);
            processor.add(1);
            var processedVal = processor.processAndGetNext();
            chai_1.expect(processedVal.isPresent()).to.equal(true);
            chai_1.expect(processedVal.get()).to.equal('1');
        });
        it('should should return empty when has no values', function () {
            var Function = function (n) { return "" + n; };
            var processor = processor_1.Processor.mapProcessor(Function);
            var processedVal = processor.processAndGetNext();
            chai_1.expect(processedVal.isPresent()).to.equal(false);
        });
        it('should not have next after getting all values', function () {
            var Function = function (n) { return "" + n; };
            var processor = processor_1.Processor.mapProcessor(Function);
            processor.add(1);
            processor.processAndGetNext();
            chai_1.expect(processor.hasNext()).to.equal(false);
        });
        it('should process in order of additions', function () {
            var Function = function (n) { return "" + n; };
            var processor = processor_1.Processor.mapProcessor(Function);
            processor.add(1);
            processor.add(2);
            var val1 = processor.processAndGetNext();
            var val2 = processor.processAndGetNext();
            chai_1.expect(val1.isPresent()).to.equal(true);
            chai_1.expect(val1.get()).to.equal('1');
            chai_1.expect(val2.isPresent()).to.equal(true);
            chai_1.expect(val2.get()).to.equal('2');
        });
        it('should process Function lazily on value fetch', function () {
            var count = 0;
            var Function = function (n) { count++; return "" + n; };
            var processor = processor_1.Processor.mapProcessor(Function);
            processor.add(1);
            processor.add(2);
            chai_1.expect(count).to.equal(0);
            var val1 = processor.processAndGetNext();
            chai_1.expect(count).to.equal(1);
            var val2 = processor.processAndGetNext();
            chai_1.expect(count).to.equal(2);
        });
    });
    describe('FilterProcessor tests', function () {
        it('should be a stateless processor', function () {
            var predicate = function (n) { return n < 10; };
            var processor = processor_1.Processor.filterProcessor(predicate);
            chai_1.expect(processor.isStateless()).to.equal(true);
        });
        it('should not have next with no values', function () {
            var predicate = function (n) { return n < 10; };
            var processor = processor_1.Processor.filterProcessor(predicate);
            chai_1.expect(processor.hasNext()).to.equal(false);
        });
        it('should have next when a value as been added', function () {
            var predicate = function (n) { return n < 10; };
            var processor = processor_1.Processor.filterProcessor(predicate);
            processor.add(1);
            chai_1.expect(processor.hasNext()).to.equal(true);
        });
        it('should run values through predicate when retrieved', function () {
            var predicate = function (n) { return n < 10; };
            var processor = processor_1.Processor.filterProcessor(predicate);
            processor.add(1);
            chai_1.expect(processor.processAndGetNext().get()).to.equal(1);
            processor.add(11);
            chai_1.expect(processor.processAndGetNext().isPresent()).to.equal(false);
        });
        it('should should return empty when has no values', function () {
            var predicate = function (n) { return n < 10; };
            var processor = processor_1.Processor.filterProcessor(predicate);
            chai_1.expect(processor.processAndGetNext().isPresent()).to.equal(false);
        });
        it('should not have next after getting all values', function () {
            var predicate = function (n) { return n < 10; };
            var processor = processor_1.Processor.filterProcessor(predicate);
            processor.add(1);
            processor.processAndGetNext();
            chai_1.expect(processor.hasNext()).to.equal(false);
        });
        it('should process in order of additions', function () {
            var predicate = function (n) { return n < 10; };
            var processor = processor_1.Processor.filterProcessor(predicate);
            processor.add(1);
            processor.add(2);
            chai_1.expect(processor.processAndGetNext().get()).to.equal(1);
            chai_1.expect(processor.processAndGetNext().get()).to.equal(2);
        });
        it('should process predicate lazily on value fetch', function () {
            var count = 0;
            var predicate = function (n) { count++; return n < 10; };
            var processor = processor_1.Processor.filterProcessor(predicate);
            processor.add(1);
            processor.add(11);
            chai_1.expect(count).to.equal(0);
            processor.processAndGetNext();
            chai_1.expect(count).to.equal(1);
            processor.processAndGetNext();
            chai_1.expect(count).to.equal(2);
        });
    });
    describe('ListFlatmapProcessor tests', function () {
        it('should be a stateless processor', function () {
            var Function = function (s) { return s.split(''); };
            var processor = processor_1.Processor.listFlatMapProcessor(Function);
            chai_1.expect(processor.isStateless()).to.equal(true);
        });
        it('should not have next with no values', function () {
            var Function = function (s) { return s.split(''); };
            var processor = processor_1.Processor.listFlatMapProcessor(Function);
            chai_1.expect(processor.hasNext()).to.equal(false);
        });
        it('should have next when a value as been added', function () {
            var Function = function (s) { return s.split(''); };
            var processor = processor_1.Processor.listFlatMapProcessor(Function);
            processor.add('cat');
            chai_1.expect(processor.hasNext()).to.equal(true);
        });
        it('should run values through Function when retrieved', function () {
            var Function = function (s) { return s.split(''); };
            var processor = processor_1.Processor.listFlatMapProcessor(Function);
            processor.add('cat');
            var val = processor.processAndGetNext();
            chai_1.expect(val.isPresent()).to.equal(true);
            chai_1.expect(val.get()).to.equal('c');
        });
        it('should hasNext be true if in the middle of a processing', function () {
            var Function = function (s) { return s.split(''); };
            var processor = processor_1.Processor.listFlatMapProcessor(Function);
            processor.add('cat');
            chai_1.expect(processor.hasNext()).to.equal(true);
            processor.processAndGetNext();
            chai_1.expect(processor.hasNext()).to.equal(true);
            processor.processAndGetNext();
            chai_1.expect(processor.hasNext()).to.equal(true);
        });
        it('should should return empty when has no values', function () {
            var Function = function (s) { return s.split(''); };
            var processor = processor_1.Processor.listFlatMapProcessor(Function);
            var val = processor.processAndGetNext();
            chai_1.expect(val.isPresent()).to.equal(false);
        });
        it('should not have next after getting all values', function () {
            it('should hasNext be true if in the middle of a processing', function () {
                var Function = function (s) { return s.split(''); };
                var processor = processor_1.Processor.listFlatMapProcessor(Function);
                processor.add('cat');
                processor.processAndGetNext();
                processor.processAndGetNext();
                processor.processAndGetNext();
                chai_1.expect(processor.hasNext()).to.equal(false);
            });
        });
        it('should process in order of additions', function () {
            var Function = function (s) { return s.split(''); };
            var processor = processor_1.Processor.listFlatMapProcessor(Function);
            processor.add('ca');
            processor.add('do');
            chai_1.expect(processor.processAndGetNext().get()).to.equal('c');
            chai_1.expect(processor.processAndGetNext().get()).to.equal('a');
            chai_1.expect(processor.processAndGetNext().get()).to.equal('d');
            chai_1.expect(processor.processAndGetNext().get()).to.equal('o');
        });
        it('should process Function lazily on value fetch', function () {
            var count = 0;
            var Function = function (s) { count++; return s.split(''); };
            var processor = processor_1.Processor.listFlatMapProcessor(Function);
            processor.add('ca');
            processor.add('do');
            chai_1.expect(count).to.equal(0);
            processor.processAndGetNext();
            chai_1.expect(count).to.equal(1);
        });
        it('should try process and return optionalEmpty if input list has null', function () {
            var Function = function (s) { return s.split(''); };
            var processor = processor_1.Processor.listFlatMapProcessor(Function);
            processor.add(null);
            chai_1.expect(processor.processAndGetNext().isPresent()).to.equal(false);
        });
    });
    describe('DistinctProcessor tests', function () {
        it('should be a stateless processor', function () {
            var comparator = function (n1, n2) { return n1 === n2; };
            var processor = processor_1.Processor.distinctProcessor(comparator);
            chai_1.expect(processor.isStateless()).to.equal(true);
        });
        it('should not have next with no values', function () {
            var comparator = function (n1, n2) { return n1 === n2; };
            var processor = processor_1.Processor.distinctProcessor(comparator);
            chai_1.expect(processor.hasNext()).to.equal(false);
        });
        it('should have next when a value as been added', function () {
            var comparator = function (n1, n2) { return n1 === n2; };
            var processor = processor_1.Processor.distinctProcessor(comparator);
            processor.add(1);
            chai_1.expect(processor.hasNext()).to.equal(true);
        });
        it('should have next when a value as been added and processing kicked off', function () {
            var comparator = function (n1, n2) { return n1 === n2; };
            var processor = processor_1.Processor.distinctProcessor(comparator);
            processor.add(1);
            processor.add(2);
            processor.processAndGetNext();
            chai_1.expect(processor.hasNext()).to.equal(true);
        });
        it('should run values through Function when retrieved', function () {
            var comparator = function (n1, n2) { return n1 === n2; };
            var processor = processor_1.Processor.distinctProcessor(comparator);
            [1, 1].forEach(function (i) { return processor.add(i); });
            var val1 = processor.processAndGetNext();
            var val2 = processor.processAndGetNext();
            chai_1.expect(val1.get()).to.equal(1);
            chai_1.expect(val2.isPresent()).to.equal(false);
        });
        it('should should return empty when has no values', function () {
            var comparator = function (n1, n2) { return n1 === n2; };
            var processor = processor_1.Processor.distinctProcessor(comparator);
            chai_1.expect(processor.processAndGetNext().isPresent()).to.equal(false);
        });
        it('should not have next after getting all values', function () {
            var comparator = function (n1, n2) { return n1 === n2; };
            var processor = processor_1.Processor.distinctProcessor(comparator);
            [1, 1].forEach(function (i) { return processor.add(i); });
            chai_1.expect(processor.hasNext()).to.equal(true);
            var val1 = processor.processAndGetNext();
            chai_1.expect(processor.hasNext()).to.equal(true);
            var val2 = processor.processAndGetNext();
            chai_1.expect(val1.get()).to.equal(1);
            chai_1.expect(val2.isPresent()).to.equal(false);
        });
        it('should return a distinct set of added items', function () {
            var comparator = function (n1, n2) { return n1 === n2; };
            var processor = processor_1.Processor.distinctProcessor(comparator);
            [1, 2, 1].forEach(function (i) { return processor.add(i); });
            var val1 = processor.processAndGetNext();
            var val2 = processor.processAndGetNext();
            var val3 = processor.processAndGetNext();
            chai_1.expect(processor.hasNext()).to.equal(false);
            chai_1.expect(val1.get()).to.equal(1);
            chai_1.expect(val2.get()).to.equal(2);
            chai_1.expect(val3.isPresent()).to.equal(false);
        });
        it('should process Function lazily on value fetch', function () {
            var count = 0;
            var comparator = function (n1, n2) { count++; return n1 === n2; };
            var processor = processor_1.Processor.distinctProcessor(comparator);
            [1, 2, 1, 2, 1, 3].forEach(function (i) { return processor.add(i); });
            chai_1.expect(count).to.equal(0);
            processor.processAndGetNext();
            processor.processAndGetNext(); //need to fetch twice before any comparison happens
            chai_1.expect(count).to.be.greaterThan(0);
        });
    });
    describe('StreamFlatMapProcessor tests', function () {
        it('should be a stateless processor', function () {
            var Function = function (s) { return stream_1.default.of(s.split('')); };
            var processor = processor_1.Processor.streamFlatMapProcessor(Function);
            chai_1.expect(processor.isStateless()).to.equal(true);
        });
        it('should not have next with no values', function () {
            var Function = function (s) { return stream_1.default.of(s.split('')); };
            var processor = processor_1.Processor.streamFlatMapProcessor(Function);
            chai_1.expect(processor.hasNext()).to.equal(false);
        });
        it('should have next when a value as been added', function () {
            var Function = function (s) { return stream_1.default.of(s.split('')); };
            var processor = processor_1.Processor.streamFlatMapProcessor(Function);
            processor.add('cat');
            chai_1.expect(processor.hasNext()).to.equal(true);
        });
        it('should run values through Function when retrieved', function () {
            var Function = function (s) { return stream_1.default.of(s.split('')); };
            var processor = processor_1.Processor.streamFlatMapProcessor(Function);
            processor.add('cat');
            var val = processor.processAndGetNext();
            chai_1.expect(val.isPresent()).to.equal(true);
            chai_1.expect(val.get()).to.equal('c');
        });
        it('should hasNext be true if in the middle of a processing', function () {
            var Function = function (s) { return stream_1.default.of(s.split('')); };
            var processor = processor_1.Processor.streamFlatMapProcessor(Function);
            processor.add('cat');
            chai_1.expect(processor.hasNext()).to.equal(true);
            processor.processAndGetNext();
            chai_1.expect(processor.hasNext()).to.equal(true);
            processor.processAndGetNext();
            chai_1.expect(processor.hasNext()).to.equal(true);
        });
        it('should should return empty when has no values', function () {
            var Function = function (s) { return stream_1.default.of(s.split('')); };
            var processor = processor_1.Processor.streamFlatMapProcessor(Function);
            var val = processor.processAndGetNext();
            chai_1.expect(val.isPresent()).to.equal(false);
        });
        it('should not have next after getting all values', function () {
            it('should hasNext be true if in the middle of a processing', function () {
                var Function = function (s) { return stream_1.default.of(s.split('')); };
                var processor = processor_1.Processor.streamFlatMapProcessor(Function);
                processor.add('cat');
                processor.processAndGetNext();
                processor.processAndGetNext();
                processor.processAndGetNext();
                chai_1.expect(processor.hasNext()).to.equal(false);
            });
        });
        it('should process in order of additions', function () {
            var Function = function (s) { return stream_1.default.of(s.split('')); };
            var processor = processor_1.Processor.streamFlatMapProcessor(Function);
            processor.add('ca');
            processor.add('do');
            chai_1.expect(processor.processAndGetNext().get()).to.equal('c');
            chai_1.expect(processor.processAndGetNext().get()).to.equal('a');
            chai_1.expect(processor.processAndGetNext().get()).to.equal('d');
            chai_1.expect(processor.processAndGetNext().get()).to.equal('o');
        });
        it('should process Function lazily on value fetch', function () {
            var count = 0;
            var Function = function (s) { count++; return stream_1.default.of(s.split('')); };
            var processor = processor_1.Processor.streamFlatMapProcessor(Function);
            processor.add('ca');
            processor.add('do');
            chai_1.expect(count).to.equal(0);
            processor.processAndGetNext();
            chai_1.expect(count).to.equal(1);
        });
        it('should try process and return optionalEmpty if input list has null', function () {
            var Function = function (s) { return stream_1.default.of(s.split('')); };
            var processor = processor_1.Processor.streamFlatMapProcessor(Function);
            processor.add(null);
            chai_1.expect(processor.processAndGetNext().isPresent()).to.equal(false);
        });
    });
    describe('PeekProcessor tests', function () {
        it('should be a stateless processor', function () {
            var consumer = function (n) { return console.log(n); };
            var processor = processor_1.Processor.peekProcessor(consumer);
            chai_1.expect(processor.isStateless()).to.equal(true);
        });
        it('should not have next with no values', function () {
            var consumer = function (n) { return console.log(n); };
            var processor = processor_1.Processor.peekProcessor(consumer);
            chai_1.expect(processor.hasNext()).to.equal(false);
        });
        it('should have next when a value as been added', function () {
            var consumer = function (n) { return console.log(n); };
            var processor = processor_1.Processor.peekProcessor(consumer);
            processor.add(1);
            chai_1.expect(processor.hasNext()).to.equal(true);
        });
        it('should run values through consumer when retrieved', function () {
            var sum = 0;
            var consumer = function (n) { return sum = sum + n; };
            var processor = processor_1.Processor.peekProcessor(consumer);
            processor.add(1);
            processor.processAndGetNext();
            chai_1.expect(sum).to.equal(1);
        });
        it('should should return empty when has no values', function () {
            var consumer = function (n) { return console.log(n); };
            var processor = processor_1.Processor.peekProcessor(consumer);
            chai_1.expect(processor.processAndGetNext().isPresent()).to.equal(false);
        });
        it('should not have next after getting all values', function () {
            var consumer = function (n) { return console.log(n); };
            var processor = processor_1.Processor.peekProcessor(consumer);
            processor.add(1);
            processor.processAndGetNext();
            chai_1.expect(processor.hasNext()).to.equal(false);
        });
        it('should process in order of additions', function () {
            var items = [];
            var consumer = function (n) { return items.push(n); };
            var processor = processor_1.Processor.peekProcessor(consumer);
            processor.add(1);
            processor.add(2);
            processor.processAndGetNext();
            processor.processAndGetNext();
            chai_1.expect(items[0]).to.eq(1);
            chai_1.expect(items[1]).to.eq(2);
        });
        it('should return unchanged values', function () {
            var consumer = function (n) { return console.log(n); };
            var processor = processor_1.Processor.peekProcessor(consumer);
            processor.add(1);
            processor.add(2);
            var v1 = processor.processAndGetNext();
            var v2 = processor.processAndGetNext();
            chai_1.expect(v1.get()).to.eq(1);
            chai_1.expect(v2.get()).to.eq(2);
        });
        it('should process Function lazily on value fetch', function () {
            var sum = 0;
            var consumer = function (n) { return sum = sum + n; };
            var processor = processor_1.Processor.peekProcessor(consumer);
            processor.add(1);
            chai_1.expect(sum).to.eq(0);
            processor.processAndGetNext();
            chai_1.expect(sum).to.equal(1);
        });
    });
    describe('OptionalFlatMapProcessor tests', function () {
        it('should be a stateless processor', function () {
            var Function = function (n) { return optional_1.default.of(n).filter(function (i) { return i < 10; }); };
            var processor = processor_1.Processor.optionalFlatMapProcessor(Function);
            chai_1.expect(processor.isStateless()).to.equal(true);
        });
        it('should not have next with no values', function () {
            var Function = function (n) { return optional_1.default.of(n).filter(function (i) { return i < 10; }); };
            var processor = processor_1.Processor.optionalFlatMapProcessor(Function);
            chai_1.expect(processor.hasNext()).to.equal(false);
        });
        it('should have next when a value as been added', function () {
            var Function = function (n) { return optional_1.default.of(n).filter(function (i) { return i < 10; }); };
            var processor = processor_1.Processor.optionalFlatMapProcessor(Function);
            processor.add(1);
            chai_1.expect(processor.hasNext()).to.equal(true);
        });
        it('should run values through Function when retrieved', function () {
            var Function = function (n) { return optional_1.default.of(n).filter(function (i) { return i < 10; }); };
            var processor = processor_1.Processor.optionalFlatMapProcessor(Function);
            processor.add(11);
            var v1 = processor.processAndGetNext();
            chai_1.expect(v1.isPresent()).to.equal(false);
        });
        it('should should return empty when has no values', function () {
            var Function = function (n) { return optional_1.default.of(n).filter(function (i) { return i < 10; }); };
            var processor = processor_1.Processor.optionalFlatMapProcessor(Function);
            var v1 = processor.processAndGetNext();
            chai_1.expect(v1.isPresent()).to.equal(false);
        });
        it('should not have next after getting all values', function () {
            var Function = function (n) { return optional_1.default.of(n).filter(function (i) { return i < 10; }); };
            var processor = processor_1.Processor.optionalFlatMapProcessor(Function);
            processor.add(11);
            var v1 = processor.processAndGetNext();
            chai_1.expect(processor.hasNext()).to.equal(false);
        });
        it('should process in order of additions', function () {
            var Function = function (n) { return optional_1.default.of(n).filter(function (i) { return i < 10; }); };
            var processor = processor_1.Processor.optionalFlatMapProcessor(Function);
            processor.add(1);
            processor.add(11);
            var v1 = processor.processAndGetNext();
            var v2 = processor.processAndGetNext();
            chai_1.expect(v1.isPresent()).to.equal(true);
            chai_1.expect(v1.get()).to.equal(1);
            chai_1.expect(v2.isPresent()).to.equal(false);
        });
        it('should get only present values from Function', function () {
            var Function = function (n) { return optional_1.default.of(n).filter(function (i) { return i < 10; }); };
            var processor = processor_1.Processor.optionalFlatMapProcessor(Function);
            processor.add(1);
            processor.add(11);
            var v1 = processor.processAndGetNext();
            var v2 = processor.processAndGetNext();
            chai_1.expect(v1.isPresent()).to.equal(true);
            chai_1.expect(v1.get()).to.equal(1);
            chai_1.expect(v2.isPresent()).to.equal(false);
        });
        it('should process Function lazily on value fetch', function () {
            var count = 0;
            var Function = function (n) { count++; return optional_1.default.of(n).filter(function (i) { return i < 10; }); };
            var processor = processor_1.Processor.optionalFlatMapProcessor(Function);
            processor.add(11);
            chai_1.expect(count).to.eq(0);
            var v1 = processor.processAndGetNext();
            chai_1.expect(count).to.eq(1);
        });
    });
    describe('SkipProcessor tests', function () {
        it('should be a stateless processor', function () {
            var processor = processor_1.Processor.skipProcessor(1);
            chai_1.expect(processor.isStateless()).to.equal(true);
        });
        it('should not have next with no values', function () {
            var processor = processor_1.Processor.skipProcessor(1);
            chai_1.expect(processor.hasNext()).to.equal(false);
        });
        it('should have next when a value as been added', function () {
            var processor = processor_1.Processor.skipProcessor(1);
            processor.add(1);
            chai_1.expect(processor.hasNext()).to.equal(true);
        });
        it('should throw out values when asked to skip', function () {
            var processor = processor_1.Processor.skipProcessor(1);
            processor.add(1);
            var v = processor.processAndGetNext();
            chai_1.expect(v.isPresent()).to.equal(false);
        });
        it('should throw out values when asked to skip, getting next', function () {
            var processor = processor_1.Processor.skipProcessor(1);
            processor.add(1);
            processor.add(2);
            var v1 = processor.processAndGetNext();
            var v2 = processor.processAndGetNext();
            chai_1.expect(v1.isPresent()).to.equal(false);
            chai_1.expect(v2.get()).to.equal(2);
        });
        it('should should return empty when has no values', function () {
            it('should throw out values when asked to skip', function () {
                var processor = processor_1.Processor.skipProcessor(1);
                var v1 = processor.processAndGetNext();
                chai_1.expect(v1.isPresent()).to.equal(false);
            });
        });
        it('should not have next after getting all values', function () {
            it('should throw out values when asked to skip', function () {
                var processor = processor_1.Processor.skipProcessor(1);
                processor.add(1);
                processor.processAndGetNext();
                chai_1.expect(processor.hasNext()).to.equal(false);
            });
        });
        it('should process in order of additions', function () {
            var processor = processor_1.Processor.skipProcessor(1);
            processor.add(1);
            processor.add(2);
            var v1 = processor.processAndGetNext();
            var v2 = processor.processAndGetNext();
            chai_1.expect(v1.isPresent()).to.equal(false);
            chai_1.expect(v2.get()).to.equal(2);
        });
    });
    describe('SortProcessor tests', function () {
        it('should not be a stateless processor', function () {
            var comparator = functions_1.Comparator.default();
            var processor = processor_1.Processor.sortProcessor(comparator);
            chai_1.expect(processor.isStateless()).to.equal(false);
        });
        it('should not be short circuiting', function () {
            var comparator = functions_1.Comparator.default();
            var processor = processor_1.Processor.sortProcessor(comparator);
            chai_1.expect(processor.isShortCircuting()).to.equal(false);
        });
        it('should not have next with no values', function () {
            var comparator = functions_1.Comparator.default();
            var processor = processor_1.Processor.sortProcessor(comparator);
            chai_1.expect(processor.hasNext()).to.equal(false);
        });
        it('should have next when a value as been added', function () {
            var comparator = functions_1.Comparator.default();
            var processor = processor_1.Processor.sortProcessor(comparator);
            processor.add(1);
            chai_1.expect(processor.isStateless()).to.equal(false);
        });
        it('should have next when a values have been processed', function () {
            var comparator = functions_1.Comparator.default();
            var processor = processor_1.Processor.sortProcessor(comparator);
            processor.add(2);
            processor.add(1);
            processor.processAndGetNext();
            chai_1.expect(processor.hasNext()).to.equal(true);
        });
        it('should run values through sorter when first processed only', function () {
            var count = 0;
            var comparator = function (n, n2) { count++; return 0; };
            var processor = processor_1.Processor.sortProcessor(comparator);
            processor.add(2);
            processor.add(1);
            chai_1.expect(count).to.eq(0);
            processor.processAndGetNext();
            var firstCount = count;
            chai_1.expect(count).to.be.greaterThan(0);
            processor.processAndGetNext();
            chai_1.expect(count).to.eq(firstCount);
        });
        it('should should return empty when has no values', function () {
            var comparator = functions_1.Comparator.default();
            var processor = processor_1.Processor.sortProcessor(comparator);
            processor.add(2);
            processor.add(1);
            processor.processAndGetNext();
            processor.processAndGetNext();
            chai_1.expect(processor.processAndGetNext().isPresent()).to.equal(false);
        });
        it('should not have next after getting all values', function () {
            var comparator = functions_1.Comparator.default();
            var processor = processor_1.Processor.sortProcessor(comparator);
            processor.add(2);
            processor.add(1);
            processor.processAndGetNext();
            processor.processAndGetNext();
            chai_1.expect(processor.hasNext()).to.equal(false);
        });
        it('should process Function lazily on value fetch', function () {
            var count = 0;
            var comparator = function (n, n2) { count++; return 0; };
            var processor = processor_1.Processor.sortProcessor(comparator);
            processor.add(2);
            processor.add(1);
            chai_1.expect(count).to.eq(0);
            processor.processAndGetNext();
            chai_1.expect(count).to.be.greaterThan(0);
        });
        it('should return values in sorted order', function () {
            var comparator = functions_1.Comparator.default();
            var processor = processor_1.Processor.sortProcessor(comparator);
            processor.add(2);
            processor.add(1);
            processor.add(4);
            chai_1.expect(processor.processAndGetNext().get()).to.equal(1);
            chai_1.expect(processor.processAndGetNext().get()).to.equal(2);
            chai_1.expect(processor.processAndGetNext().get()).to.equal(4);
        });
    });
    describe('LimitProcessor tests', function () {
        it('should not be a stateless processor', function () {
            var processor = processor_1.Processor.limitProcessor(1);
            chai_1.expect(processor.isStateless()).to.equal(true);
        });
        it('should not be a short circuiting processor', function () {
            var processor = processor_1.Processor.limitProcessor(1);
            chai_1.expect(processor.isShortCircuting()).to.equal(true);
        });
        it('has next while limit not yet reached, even when empty', function () {
            var processor = processor_1.Processor.limitProcessor(1);
            chai_1.expect(processor.hasNext()).to.equal(true);
        });
        it('has next while limit not yet reached', function () {
            var processor = processor_1.Processor.limitProcessor(1);
            processor.add(1);
            chai_1.expect(processor.hasNext()).to.equal(true);
        });
        it('should run values through Function when retrieved', function () {
            var processor = processor_1.Processor.limitProcessor(1);
            processor.add(1);
            processor.add(2);
            chai_1.expect(processor.processAndGetNext().get()).to.equal(1);
            chai_1.expect(processor.processAndGetNext().isPresent()).to.equal(false);
        });
        it('should run values through Function when retrieved', function () {
            var processor = processor_1.Processor.limitProcessor(1);
            processor.add(1);
            processor.add(1);
            processor.add(1);
            chai_1.expect(processor.processAndGetNext().isPresent()).to.equal(true);
            chai_1.expect(processor.processAndGetNext().isPresent()).to.equal(false);
        });
        it('should not have next after getting the number specified by limit', function () {
            var processor = processor_1.Processor.limitProcessor(1);
            processor.add(1);
            processor.add(2);
            processor.add(2);
            processor.add(2);
            processor.processAndGetNext();
            processor.processAndGetNext();
            chai_1.expect(processor.hasNext()).to.equal(false);
        });
    });
});
