"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var chai_1 = require("chai");
var spies = require("chai-spies");
var source_1 = require("../source");
var stream_1 = require("../stream");
chai_1.use(spies);
describe('Source tests', function () {
    describe('Iterate Source', function () {
        it('should return seed first', function () {
            var seed = 0;
            var Function = function (n) { return n + 1; };
            var source = source_1.Source.iterateSource(seed, Function);
            var result = source.get();
            chai_1.expect(result).to.equal(seed);
        });
        it('should return seed first', function () {
            var seed = 0;
            var Function = function (n) { return n + 1; };
            var source = source_1.Source.iterateSource(seed, Function);
            var result = source.get();
            chai_1.expect(result).to.equal(seed);
        });
        it('should transform seed each call', function () {
            var seed = 0;
            var Function = function (n) { return n + 1; };
            var source = source_1.Source.iterateSource(seed, Function);
            chai_1.expect(source.get()).to.equal(seed);
            chai_1.expect(source.get()).to.equal(Function(seed));
            chai_1.expect(source.get()).to.equal(Function(Function(seed)));
        });
        it('should hasNext should be true', function () {
            var Function = function (n) { return n + 1; };
            var source = source_1.Source.iterateSource(0, Function);
            chai_1.expect(source.hasNext()).to.equal(true);
        });
    });
    describe('Supplier Source', function () {
        it('should return value from supplier', function () {
            var value = 0;
            var supplier = function () { return value; };
            var source = source_1.Source.supplierSource(supplier);
            var result = source.get();
            chai_1.expect(result).to.equal(value);
        });
        it('should call supplier', function () {
            var value = 0;
            var supplier = function () { return value; };
            var supplierSpy = chai_1.spy(supplier);
            var source = source_1.Source.supplierSource(supplierSpy);
            source.get();
            chai_1.expect(supplierSpy).to.have.been.called();
        });
        it('should hasNext should be true', function () {
            var supplier = function () { return 0; };
            var source = source_1.Source.supplierSource(supplier);
            chai_1.expect(source.hasNext()).to.equal(true);
        });
    });
    describe('Array Source', function () {
        it('should return values from array', function () {
            var sourceArray = [1, 2, 3];
            var source = source_1.Source.arraySource(sourceArray);
            chai_1.expect(source.get()).to.equal(sourceArray[0]);
            chai_1.expect(source.get()).to.equal(sourceArray[1]);
            chai_1.expect(source.get()).to.equal(sourceArray[2]);
        });
        it('should hasNext when still values in array', function () {
            var sourceArray = [1, 2, 3];
            var source = source_1.Source.arraySource(sourceArray);
            chai_1.expect(source.get()).to.equal(sourceArray[0]);
            chai_1.expect(source.get()).to.equal(sourceArray[1]);
            chai_1.expect(source.hasNext()).to.equal(true);
        });
        it('should hasNext when still values in array', function () {
            var sourceArray = [1, 2];
            var source = source_1.Source.arraySource(sourceArray);
            chai_1.expect(source.get()).to.equal(sourceArray[0]);
            chai_1.expect(source.get()).to.equal(sourceArray[1]);
            chai_1.expect(source.hasNext()).to.equal(false);
        });
    });
    describe('Concat Source', function () {
        it('should hasNext true, first stream has values, second does not', function () {
            var stream1 = stream_1.default.ofValues(1, 2, 3);
            var stream2 = stream_1.default.ofValues();
            var source = source_1.Source.concatSource(stream1, stream2);
            chai_1.expect(source.hasNext()).to.equal(true);
        });
        it('should hasNext first stream empty, second not empty', function () {
            var stream2 = stream_1.default.ofValues(1, 2, 3);
            var stream1 = stream_1.default.ofValues();
            var source = source_1.Source.concatSource(stream1, stream2);
            chai_1.expect(source.hasNext()).to.equal(true);
        });
        it('should hasNext true when both have values', function () {
            var stream1 = stream_1.default.ofValues(1, 2, 3);
            var stream2 = stream_1.default.ofValues(4, 5, 6);
            var source = source_1.Source.concatSource(stream1, stream2);
            chai_1.expect(source.hasNext()).to.equal(true);
        });
        it('should hasNext fa;se when both have no values', function () {
            var stream1 = stream_1.default.ofValues();
            var stream2 = stream_1.default.ofValues();
            var source = source_1.Source.concatSource(stream1, stream2);
            chai_1.expect(source.hasNext()).to.equal(false);
        });
        it('should get values first from stream1', function () {
            var stream1 = stream_1.default.ofValues(1);
            var stream2 = stream_1.default.ofValues(2);
            var source = source_1.Source.concatSource(stream1, stream2);
            var result = source.get();
            chai_1.expect(result.get()).to.equal(1);
        });
        it('should get values from stream2 after stream1 empty', function () {
            var stream1 = stream_1.default.ofValues(1);
            var stream2 = stream_1.default.ofValues(2);
            var source = source_1.Source.concatSource(stream1, stream2);
            source.get();
            var result = source.get();
            chai_1.expect(result.get()).to.equal(2);
        });
    });
    describe('Range Source', function () {
        it('return the start value first', function () {
            var source = source_1.Source.rangeSource(0, 4);
            chai_1.expect(source.get()).to.equal(0);
        });
        it('if start and end are same, empty source', function () {
            var source = source_1.Source.rangeSource(0, 0);
            chai_1.expect(source.hasNext()).to.equal(false);
        });
        it('default step to 1 when none supplied, ancending', function () {
            var source = source_1.Source.rangeSource(1, 3);
            var first = source.get();
            var second = source.get();
            var step = second - first;
            chai_1.expect(Math.abs(step)).to.equal(1);
        });
        it('default step to 1 when none supplied, descending', function () {
            var source = source_1.Source.rangeSource(-1, -3);
            var first = source.get();
            var second = source.get();
            var step = second - first;
            chai_1.expect(Math.abs(step)).to.equal(1);
        });
        it('uses supplied step as absolute value', function () {
            var source = source_1.Source.rangeSource(1, 10, 2);
            var first = source.get();
            var second = source.get();
            var step = second - first;
            chai_1.expect(Math.abs(step)).to.equal(2);
        });
        it('uses supplied step as absolute value', function () {
            var source = source_1.Source.rangeSource(1, 10, -2);
            var first = source.get();
            var second = source.get();
            var step = second - first;
            chai_1.expect(Math.abs(step)).to.equal(2);
        });
        it('uses supplied step as absolute value, decending range', function () {
            var source = source_1.Source.rangeSource(-1, -10, 2);
            var first = source.get();
            var second = source.get();
            var step = second - first;
            chai_1.expect(Math.abs(step)).to.equal(2);
        });
        it('uses supplied step as absolute value, descending range', function () {
            var source = source_1.Source.rangeSource(-1, -10, -2);
            var first = source.get();
            var second = source.get();
            var step = second - first;
            chai_1.expect(Math.abs(step)).to.equal(2);
        });
        it('should exclude end value', function () {
            var source = source_1.Source.rangeSource(0, 1);
            var first = source.get();
            var second = source.get();
            chai_1.expect(first).to.equal(0);
            chai_1.expect(second).to.equal(undefined);
        });
        it('should not hasNext when end values is next', function () {
            var source = source_1.Source.rangeSource(0, 1);
            source.get();
            chai_1.expect(source.hasNext()).to.equal(false);
        });
    });
});
