"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var stream_1 = require("../stream");
var collectors_1 = require("../collectors");
var chai_1 = require("chai");
var spies = require("chai-spies");
var optional_1 = require("../optional");
var functions_1 = require("../functions");
chai_1.use(spies);
var testItemComparator = function (i1, i2) { return functions_1.Comparator.default()(i1.i, i2.i); };
describe('Stream tests', function () {
    describe('collect to list tests', function () {
        it('it should put all elements in the stream into the collector', function () {
            var source = [1, 2, 3, 4, 5];
            var stream = stream_1.default.of(source);
            var result = stream.collect(collectors_1.default.toList());
            chai_1.expect(result.length).to.equal(5);
            source.forEach(function (item) { return chai_1.expect(result).to.contain(item); });
        });
        it('it should handle empty stream', function () {
            var stream = stream_1.default.empty();
            var result = stream.collect(collectors_1.default.toList());
            chai_1.expect(result.length).to.equal(0);
        });
    });
    describe('Stream.concat', function () {
        it('it concatenate two streams with values', function () {
            var s1 = stream_1.default.ofValues(1, 2, 3);
            var s2 = stream_1.default.ofValues(4, 5, 6);
            var concat = stream_1.default.concat(s1, s2);
            var result = concat.toArray();
            chai_1.expect(result.length).to.equal(6);
            chai_1.expect(result[0]).to.equal(1);
            chai_1.expect(result[5]).to.equal(6);
        });
        it('it concatenate first stream empty', function () {
            var s1 = stream_1.default.empty();
            var s2 = stream_1.default.ofValues(4, 5, 6);
            var concat = stream_1.default.concat(s1, s2);
            var result = concat.toArray();
            chai_1.expect(result.length).to.equal(3);
            chai_1.expect(result[0]).to.equal(4);
            chai_1.expect(result[2]).to.equal(6);
        });
        it('it concatenate second stream empty', function () {
            var s1 = stream_1.default.ofValues(1, 2, 3);
            var s2 = stream_1.default.empty();
            var concat = stream_1.default.concat(s1, s2);
            var result = concat.toArray();
            chai_1.expect(result.length).to.equal(3);
            chai_1.expect(result[0]).to.equal(1);
            chai_1.expect(result[2]).to.equal(3);
        });
        it('it concatenate two complex streams', function () {
            var s1 = stream_1.default.ofValues('1', '2', '3').map(function (i) { return parseInt(i); }).distinct();
            var s2 = stream_1.default.ofValues(4, 5, 6, 10, 20, 50).distinct().filter(function (i) { return i < 10; });
            var concat = stream_1.default.concat(s1, s2);
            var result = concat.toArray();
            chai_1.expect(result.length).to.equal(6);
            chai_1.expect(result[0]).to.equal(1);
            chai_1.expect(result[5]).to.equal(6);
        });
        it('it concatenate two complex streams with infinite', function () {
            var s1 = stream_1.default.generate(function () { return 1; }).limit(3);
            var s2 = stream_1.default.ofValues(4, 5, 6, 10, 20, 50).distinct().filter(function (i) { return i < 10; });
            var concat = stream_1.default.concat(s1, s2);
            var result = concat.toArray();
            chai_1.expect(result.length).to.equal(6);
            chai_1.expect(result[0]).to.equal(1);
            chai_1.expect(result[5]).to.equal(6);
        });
        it('it should handle empty stream', function () {
            var stream = stream_1.default.empty();
            var result = stream.collect(collectors_1.default.toList());
            chai_1.expect(result.length).to.equal(0);
        });
    });
    describe('forEachOrdered', function () {
        it('it should run consumer on all elements', function () {
            var sum = 0;
            var source = [1, 1, 1, 1, 1];
            var stream = stream_1.default.of(source);
            stream.forEachOrdered(function (i) { return sum = sum + i; });
            chai_1.expect(sum).to.equal(5);
        });
        it('it should not run consumer if stream is empty', function () {
            var consumerRan = false;
            var stream = stream_1.default.empty();
            stream.forEachOrdered(function (i) { return consumerRan = true; });
            chai_1.expect(consumerRan).to.equal(false);
        });
    });
    describe('distinct', function () {
        it('it return a stream of distinct values', function () {
            var source = [4, 1, 2, 3, 2, 1, 1, 1, 1, 1];
            var expectedValues = [1, 2, 3, 4];
            var stream = stream_1.default.of(source);
            var result = stream.distinct().collect(collectors_1.default.toList());
            chai_1.expect(result.length).to.equal(4);
            expectedValues.forEach(function (item) { return chai_1.expect(result).to.contain(item); });
        });
        it('it should work as the first intermediate operation', function () {
            var source = [4, 1, 2, 3, 2, 1, 1, 1, 1, 1];
            var expectedValues = ['1', '2', '3', '4'];
            var stream = stream_1.default.of(source);
            var result = stream.distinct().map(function (n) { return "" + n; }).collect(collectors_1.default.toList());
            chai_1.expect(result.length).to.equal(4);
            expectedValues.forEach(function (item) { return chai_1.expect(result).to.contain(item); });
        });
        it('it should work preceeded by stateless op', function () {
            it('it should work as the first intermediate operation', function () {
                var source = [4, 1, 2, 3, 2, 1, 1, 1, 1, 1];
                var expectedValues = ['1', '2', '3', '4'];
                var stream = stream_1.default.of(source);
                var result = stream
                    .map(function (n) { return "" + n; })
                    .distinct()
                    .collect(collectors_1.default.toList());
                chai_1.expect(result.length).to.equal(4);
                expectedValues.forEach(function (item) { return chai_1.expect(result).to.contain(item); });
            });
        });
        it('it should work if multiple calls in one stream', function () {
            var source = ['cat', 'bat', 'cat'];
            var expectedValues = ['c', 'a', 't', 'b'];
            var stream = stream_1.default.of(source);
            var result = stream
                .distinct()
                .flatMapList(function (word) { return word.split(''); })
                .distinct()
                .collect(collectors_1.default.toList());
            chai_1.expect(result.length).to.equal(4);
            expectedValues.forEach(function (item) { return chai_1.expect(result).to.contain(item); });
        });
        it('it should work in complicated pipeline', function () {
            var source = ['cat', 'bat', 'cat'];
            var expectedValues = ['a'];
            var stream = stream_1.default.of(source);
            var result = stream
                .distinct()
                .flatMapList(function (word) { return word.split(''); })
                .distinct()
                .filter(function (i) { return i === 'a'; })
                .collect(collectors_1.default.toList());
            chai_1.expect(result.length).to.equal(1);
            expectedValues.forEach(function (item) { return chai_1.expect(result).to.contain(item); });
        });
        it('it should work if preceeded by short circuiting op', function () {
            var source = ['abc', '123'];
            var expected = ['a', 'b'];
            var stream = stream_1.default.of(source);
            var result = stream.flatMapList(function (word) { return word.split(''); })
                .limit(2)
                .distinct()
                .collect(collectors_1.default.toList());
            chai_1.expect(result.length).to.equal(2);
            expected.forEach(function (item) { return chai_1.expect(result).to.contain(item); });
        });
        it('it should work if followed by short circuiting op', function () {
            var source = ['abc', '123'];
            var expected = ['a', 'b'];
            var stream = stream_1.default.of(source);
            var result = stream.flatMapList(function (word) { return word.split(''); })
                .distinct()
                .limit(2)
                .collect(collectors_1.default.toList());
            chai_1.expect(result.length).to.equal(2);
            expected.forEach(function (item) { return chai_1.expect(result).to.contain(item); });
        });
    });
    describe('filter tests', function () {
        it('it keeps matching values in the stream', function () {
            var source = [1, 2, 3, 11, 12, 13];
            var expectedValues = [1, 2, 3];
            var stream = stream_1.default.of(source);
            var result = stream.filter(function (i) { return i < 10; })
                .collect(collectors_1.default.toList());
            chai_1.expect(result.length).to.equal(3);
            expectedValues.forEach(function (item) { return chai_1.expect(result).to.contain(item); });
        });
        it('it should lazily filter only when terminal is called', function () {
            var count = 0;
            var source = [1, 2, 3, 11, 12, 13];
            var stream = stream_1.default.of(source);
            stream = stream.filter(function (i) { count++; return i < 10; });
            chai_1.expect(count).to.equal(0);
            stream.collect(collectors_1.default.toList());
            chai_1.expect(count).to.equal(6);
        });
    });
    describe('allMatch', function () {
        it('it should return true if all items match', function () {
            var source = [1, 2, 3];
            var stream = stream_1.default.of(source);
            var result = stream.allMatch(function (i) { return i < 5; });
            chai_1.expect(result).to.equal(true);
        });
        it('it should return false if not all items match', function () {
            var source = [1, 2, 3, 9];
            var stream = stream_1.default.of(source);
            var result = stream.allMatch(function (i) { return i < 5; });
            chai_1.expect(result).to.equal(false);
        });
        it('it should short circuit and only go until false is reached', function () {
            var source = [1, 2, 1, 1];
            var stream = stream_1.default.of(source);
            var spyMatcher = chai_1.spy(function (i) { return i === 1; });
            var result = stream.allMatch(spyMatcher);
            chai_1.expect(spyMatcher).to.be.called.exactly(2);
        });
    });
    describe('anyMatch', function () {
        it('it should return true if any items match', function () {
            var source = [1, 2, 3];
            var stream = stream_1.default.of(source);
            var result = stream.anyMatch(function (i) { return i === 2; });
            chai_1.expect(result).to.equal(true);
        });
        it('it should return false if no items match', function () {
            var source = [1, 2, 3, 9];
            var stream = stream_1.default.of(source);
            var result = stream.anyMatch(function (i) { return i === 0; });
            chai_1.expect(result).to.equal(false);
        });
        it('it should short circuit and only go until match is reached', function () {
            var source = [1, 2, 1, 1];
            var stream = stream_1.default.of(source);
            var spyMatcher = chai_1.spy(function (i) { return i === 2; });
            var result = stream.anyMatch(spyMatcher);
            chai_1.expect(spyMatcher).to.be.called.exactly(2);
        });
    });
    describe('count', function () {
        it('it should return number of values in the stream', function () {
            var source = [1, 2, 3, 11, 12, 13];
            var stream = stream_1.default.of(source);
            var result = stream.count();
            chai_1.expect(result).equal(source.length);
        });
        it('it should return number of values in the stream 0 if empty', function () {
            var stream = stream_1.default.empty();
            var result = stream.count();
            chai_1.expect(result).equal(0);
        });
        it('it should return number of values in the stream 0 if empty', function () {
            var stream = stream_1.default.ofValues(1, 1, 1);
            var result = stream.filter(function (i) { return i === 2; }).count();
            chai_1.expect(result).equal(0);
        });
    });
    describe('customCollect', function () {
        it('it should apply a mutable reduction, works for basic list builder', function () {
            var supplier = function () { return []; };
            var accumulator = function (s, n) { return s.push(n); };
            var combiner = function (s1, s2) { return s1.concat(s2); };
            var result = stream_1.default.ofValues(1, 2, 3).customCollect(supplier, accumulator, combiner);
            chai_1.expect(result.length).to.eq(3);
            chai_1.expect(result).to.contain(1);
            chai_1.expect(result).to.contain(2);
            chai_1.expect(result).to.contain(3);
        });
    });
    describe('findFirst', function () {
        it('it return the first element in the stream', function () {
            var source = [1, 2, 3, 11, 12, 13];
            var stream = stream_1.default.of(source);
            var result = stream.findFirst();
            chai_1.expect(result.isPresent()).to.equal(true);
            chai_1.expect(result.get()).to.equal(1);
        });
        it('it return optional empty if stream is empty', function () {
            var stream = stream_1.default.empty();
            var result = stream.findFirst();
            chai_1.expect(result.isPresent()).to.equal(false);
        });
        it('it should be short circuiting', function () {
            var source = [1, 2, 3, 11, 12, 13];
            var stream = stream_1.default.of(source);
            var spyTransformer = chai_1.spy(functions_1.Function.identity());
            stream.map(spyTransformer).findFirst();
            chai_1.expect(spyTransformer).to.be.called.exactly(1);
        });
    });
    describe('findFirst', function () {
        it('it return the first element in the stream', function () {
            var source = [1, 2, 3, 11, 12, 13];
            var stream = stream_1.default.of(source);
            var result = stream.findFirst();
            chai_1.expect(result.isPresent()).to.equal(true);
            chai_1.expect(result.get()).to.equal(1);
        });
        it('it return optional empty if stream is empty', function () {
            var stream = stream_1.default.empty();
            var result = stream.findFirst();
            chai_1.expect(result.isPresent()).to.equal(false);
        });
        it('it should be short circuiting', function () {
            var source = [1, 2, 3, 11, 12, 13];
            var stream = stream_1.default.of(source);
            var spyTransformer = chai_1.spy(functions_1.Function.identity());
            stream.map(spyTransformer).findFirst();
            chai_1.expect(spyTransformer).to.be.called.exactly(1);
        });
    });
    describe('findAny', function () {
        it('it return an item in the stream', function () {
            var source = [1, 2, 3, 11, 12, 13];
            var stream = stream_1.default.of(source);
            var result = stream.findAny();
            chai_1.expect(result.isPresent()).to.equal(true);
            chai_1.expect(source).to.contain(result.get());
        });
        it('it return optional empty if stream is empty', function () {
            var stream = stream_1.default.empty();
            var result = stream.findAny();
            chai_1.expect(result.isPresent()).to.equal(false);
        });
        it('it should be short circuiting', function () {
            var source = [1, 2, 3, 11, 12, 13];
            var stream = stream_1.default.of(source);
            var spyTransformer = chai_1.spy(functions_1.Function.identity());
            stream.map(spyTransformer).findAny();
            chai_1.expect(spyTransformer).to.be.called.exactly(1);
        });
    });
    describe('flatMap', function () {
        it('it should, flatten a stream of streams', function () {
            var stream1 = stream_1.default.ofValues(1, 2, 3);
            var stream2 = stream_1.default.ofValues(1, 2, 3);
            var stream = stream_1.default.ofValues(stream1, stream2);
            var listResult = stream.flatMap(function (i) { return i; }).collect(collectors_1.default.toList());
            chai_1.expect(listResult.length).to.equal(6);
            chai_1.expect(listResult[0]).to.equal(1);
            chai_1.expect(listResult[5]).to.equal(3);
        });
        it('it should lazily pull from input streams', function () {
            var s1Spy = chai_1.spy(functions_1.Function.identity());
            var s2Spy = chai_1.spy(functions_1.Function.identity());
            var stream1 = stream_1.default.ofValues(1, 2, 3).map(s1Spy);
            var stream2 = stream_1.default.ofValues(1, 2, 3).map(s2Spy);
            var stream = stream_1.default.ofValues(stream1, stream2);
            stream.flatMap(function (i) { return i; }).filter(function (i) { return i === 2; }).findFirst();
            chai_1.expect(s1Spy).to.be.called.exactly(2);
            chai_1.expect(s2Spy).to.be.called.exactly(0);
        });
    });
    describe('flatMapList', function () {
        it('it should, flatten a stream of lists', function () {
            var l1 = [1, 2, 3];
            var l2 = [1, 2, 3];
            var stream = stream_1.default.ofValues(l1, l2);
            var listResult = stream.flatMapList(function (i) { return i; }).collect(collectors_1.default.toList());
            chai_1.expect(listResult.length).to.equal(6);
            chai_1.expect(listResult[0]).to.equal(1);
            chai_1.expect(listResult[5]).to.equal(3);
        });
    });
    describe('flatMapOptional', function () {
        it('it should return stream of values when optional has a value', function () {
            var stream = stream_1.default.ofValues(optional_1.default.of(1), optional_1.default.of(2), optional_1.default.empty(), optional_1.default.of(3));
            var listResult = stream.flatMapOptional(function (i) { return i; }).collect(collectors_1.default.toList());
            chai_1.expect(listResult.length).to.equal(3);
            chai_1.expect(listResult[0]).to.equal(1);
            chai_1.expect(listResult[1]).to.equal(2);
            chai_1.expect(listResult[2]).to.equal(3);
        });
    });
    describe('limit', function () {
        it('it should reduce the size of the stream to given size', function () {
            var limit = 2;
            var source = [1, 2, 3, 11, 12, 13];
            var stream = stream_1.default.of(source);
            var result = stream.limit(limit).collect(collectors_1.default.toList());
            chai_1.expect(result.length).to.equal(limit);
        });
        it('it should not break if stream is smaller than limit', function () {
            var limit = 10;
            var source = [1, 2, 3, 11, 12, 13];
            var stream = stream_1.default.of(source);
            var result = stream.limit(limit).collect(collectors_1.default.toList());
            chai_1.expect(result.length).to.equal(source.length);
        });
    });
    describe('map', function () {
        it('it should transform all values in the stream with the given Function', function () {
            var source = [1, 2, 3];
            var expectedValues = [2, 4, 6];
            var stream = stream_1.default.of(source);
            var result = stream.map(function (i) { return i * 2; }).collect(collectors_1.default.toList());
            chai_1.expect(result.length).equal(3);
            result.forEach(function (v) { return chai_1.expect(expectedValues).to.contain(v); });
        });
        it('it should lazily transform', function () {
            var source = [1, 2, 3];
            var stream = stream_1.default.of(source);
            var spyTransfomer = chai_1.spy(functions_1.Function.identity());
            stream.map(spyTransfomer).findFirst();
            chai_1.expect(spyTransfomer).to.be.called.exactly(1);
        });
    });
    describe('max', function () {
        it('find the max value based on default comparator', function () {
            var source = [1, 2, 3, 11, 12, 13];
            var stream = stream_1.default.of(source);
            var max = stream.max();
            chai_1.expect(max.isPresent()).equal(true);
            chai_1.expect(max.get()).equal(13);
        });
        it('return optional empty if stream is empty', function () {
            var stream = stream_1.default.empty();
            var max = stream.max();
            chai_1.expect(max.isPresent()).equal(false);
        });
        it('return max value based on custom comparator', function () {
            var source = [{ i: 1 }, { i: 10 }, { i: 5 }];
            var max = stream_1.default.of(source).max(testItemComparator);
            chai_1.expect(max.isPresent()).equal(true);
            chai_1.expect(max.get().i).equal(10);
        });
    });
    describe('min', function () {
        it('find the min value based on default comparator', function () {
            var source = [1, 2, 3, 11, 12, 13];
            var stream = stream_1.default.of(source);
            var min = stream.min();
            chai_1.expect(min.isPresent()).equal(true);
            chai_1.expect(min.get()).equal(1);
        });
        it('return optional empty if stream is empty', function () {
            var stream = stream_1.default.empty();
            var min = stream.min();
            chai_1.expect(min.isPresent()).equal(false);
        });
        it('return min value based on custom comparator', function () {
            var source = [{ i: 1 }, { i: 10 }, { i: 5 }];
            var min = stream_1.default.of(source).min(testItemComparator);
            chai_1.expect(min.isPresent()).equal(true);
            chai_1.expect(min.get().i).equal(1);
        });
    });
    describe('noneMatch', function () {
        it('it should return true if no items match', function () {
            var source = [1, 2, 3];
            var stream = stream_1.default.of(source);
            var result = stream.noneMatch(function (i) { return i === 4; });
            chai_1.expect(result).to.equal(true);
        });
        it('it should return false if any items match', function () {
            var source = [1, 2, 3, 9];
            var stream = stream_1.default.of(source);
            var result = stream.noneMatch(function (i) { return i === 2; });
            chai_1.expect(result).to.equal(false);
        });
        it('it should short circuit and only go until match is reached', function () {
            var source = [1, 2, 1, 1];
            var stream = stream_1.default.of(source);
            var spyMatcher = chai_1.spy(function (i) { return i === 2; });
            stream.noneMatch(spyMatcher);
            chai_1.expect(spyMatcher).to.be.called.exactly(2);
        });
    });
    describe('peek', function () {
        it('it should return the values given', function () {
            var source = [1, 2, 3];
            var stream = stream_1.default.of(source);
            var result = stream.peek(functions_1.Consumer.sink()).collect(collectors_1.default.toList());
            chai_1.expect(result.length).equal(3);
            result.forEach(function (v) { return chai_1.expect(source).to.contain(v); });
        });
        it('it call consumer on values', function () {
            var stream = stream_1.default.ofValue(1);
            var spyConsumer = chai_1.spy(functions_1.Consumer.sink());
            stream.peek(spyConsumer).findFirst();
            chai_1.expect(spyConsumer).to.be.called.with(1);
        });
    });
    describe('reduce', function () {
        it('it should return empty if stream is empty and no initial is specified', function () {
            var stream = stream_1.default.empty();
            var result = stream.reduce(function (x, y) { return x + y; });
            chai_1.expect(result.isPresent()).equal(false);
        });
        it('it should return initial value in optional if stream is empty and initial is given', function () {
            var initial = 1;
            var stream = stream_1.default.empty();
            var result = stream.reduce(function (x, y) { return x + y; }, initial);
            chai_1.expect(result.isPresent()).equal(true);
            chai_1.expect(result.get()).to.equal(initial);
        });
        it('it apply the reduction function on the elements of the stream, when no initial value is given', function () {
            var stream = stream_1.default.ofValues(1, 1, 1);
            var result = stream.reduce(function (x, y) { return x + y; });
            chai_1.expect(result.isPresent()).equal(true);
            chai_1.expect(result.get()).to.equal(3);
        });
        it('it apply the reduction function on the elements of the stream, when initial is given', function () {
            var initial = 1;
            var stream = stream_1.default.ofValues(1, 1, 1);
            var result = stream.reduce(function (x, y) { return x + y; }, initial);
            chai_1.expect(result.isPresent()).equal(true);
            chai_1.expect(result.get()).to.equal(4);
        });
    });
    describe('skip', function () {
        it('it should, remove n number of elements from the stream', function () {
            var stream = stream_1.default.ofValues(1, 2, 3, 4);
            var result = stream.skip(2).collect(collectors_1.default.toList());
            chai_1.expect(result.length).equal(2);
            chai_1.expect(result[0]).to.equal(3);
        });
        it('it should, do nothing if given 0', function () {
            var stream = stream_1.default.ofValues(1, 2, 3, 4);
            var result = stream.skip(0).collect(collectors_1.default.toList());
            chai_1.expect(result.length).equal(4);
            chai_1.expect(result[0]).to.equal(1);
        });
        it('it should, do nothing if given negative value', function () {
            var stream = stream_1.default.ofValues(1, 2, 3, 4);
            var result = stream.skip(-5).collect(collectors_1.default.toList());
            chai_1.expect(result.length).equal(4);
            chai_1.expect(result[0]).to.equal(1);
        });
        it('it should, make stream empty if no elements are left', function () {
            var stream = stream_1.default.ofValues(1, 2, 3, 4);
            var result = stream.skip(5).collect(collectors_1.default.toList());
            chai_1.expect(result.length).equal(0);
        });
    });
    describe('sorted', function () {
        it('it should, return a sorted stream based on default sort', function () {
            var stream = stream_1.default.ofValues(2, 1, 4, 3);
            var result = stream.sorted().collect(collectors_1.default.toList());
            chai_1.expect(result.length).equal(4);
            chai_1.expect(result[0]).to.equal(1);
            chai_1.expect(result[1]).to.equal(2);
            chai_1.expect(result[2]).to.equal(3);
            chai_1.expect(result[3]).to.equal(4);
        });
        it('it should, return a sorted stream based on reverse sorter', function () {
            var stream = stream_1.default.ofValues(2, 1, 4, 3);
            var result = stream.sorted(function (n1, n2) { return functions_1.Comparator.default()(n2, n1); }).collect(collectors_1.default.toList());
            chai_1.expect(result.length).equal(4);
            chai_1.expect(result[0]).to.equal(4);
            chai_1.expect(result[1]).to.equal(3);
            chai_1.expect(result[2]).to.equal(2);
            chai_1.expect(result[3]).to.equal(1);
        });
        it('it should, return a sorted stream based on a custom sort sort', function () {
            var stream = stream_1.default.ofValues({ i: 2 }, { i: 1 }, { i: 3 }, { i: 4 });
            var result = stream.sorted(testItemComparator).collect(collectors_1.default.toList());
            chai_1.expect(result.length).equal(4);
            chai_1.expect(result[0].i).to.equal(1);
            chai_1.expect(result[1].i).to.equal(2);
            chai_1.expect(result[2].i).to.equal(3);
            chai_1.expect(result[3].i).to.equal(4);
        });
        it('it should greedily pull in values', function () {
            var mapper = chai_1.spy(function (n) { return { i: n }; });
            var stream = stream_1.default.ofValues(1, 3, 4, 2).map(mapper);
            stream.sorted(testItemComparator).findFirst();
            chai_1.expect(mapper).to.have.been.called.exactly(4);
        });
    });
    describe('toArray', function () {
        it('it should return an array representation of the stream elements', function () {
            var result = stream_1.default.ofValues(1, 2).toArray();
            chai_1.expect(result).to.be.a('Array');
            chai_1.expect(result).to.have.ordered.members([1, 2]);
        });
    });
    describe('of', function () {
        it('it should return a stream of elements from given array', function () {
            var source = [1, 2, 3];
            var stream = stream_1.default.of(source);
            var result = stream.toArray();
            chai_1.expect(result.length).to.eq(source.length);
            chai_1.expect(result).to.contain.ordered.members(source);
        });
    });
    describe('ofValues', function () {
        it('it should return a stream of elements from given array', function () {
            var source = [1, 2, 3];
            var stream = stream_1.default.ofValues.apply(stream_1.default, source);
            var result = stream.toArray();
            chai_1.expect(result.length).to.eq(source.length);
            chai_1.expect(result).to.contain.ordered.members(source);
        });
    });
    describe('ofValue', function () {
        it('it should return a stream of a single element from given array', function () {
            var source = 1;
            var stream = stream_1.default.ofValue(source);
            var result = stream.toArray();
            chai_1.expect(result.length).to.eq(1);
            chai_1.expect(result[0]).eq(source);
        });
    });
    describe('empty', function () {
        it('it should return a empty stream of elements', function () {
            var stream = stream_1.default.empty();
            var result = stream.toArray();
            chai_1.expect(result.length).to.eq(0);
        });
    });
    describe('generate', function () {
        it('it should return a stream created by a supplier function', function () {
            var supplierSpy = chai_1.spy(function () { return 1; });
            var stream = stream_1.default.generate(supplierSpy);
            var result = stream.limit(3).toArray();
            chai_1.expect(result.length).to.eq(3);
            chai_1.expect(supplierSpy).to.have.been.called.exactly(3);
        });
    });
    describe('iterate', function () {
        it('it should start stream with seed', function () {
            var seed = 0;
            var transformerSpy = chai_1.spy(function (n) { return n + 1; });
            var stream = stream_1.default.iterate(seed, transformerSpy);
            var result = stream.limit(3).toArray();
            chai_1.expect(result[0]).to.equal(seed);
        });
        it('it should start based on seed', function () {
            var seed = 0;
            var transformerSpy = chai_1.spy(function (n) { return n + 1; });
            var stream = stream_1.default.iterate(seed, transformerSpy);
            var result = stream.limit(3).toArray();
            chai_1.expect(result[0]).to.equal(seed);
            chai_1.expect(result[1]).to.equal(transformerSpy(seed));
            chai_1.expect(result[2]).to.equal(transformerSpy(transformerSpy(seed)));
        });
    });
    describe('range', function () {
        it('it should start with start value', function () {
            var start = 0;
            var end = 11;
            var step = 1;
            var stream = stream_1.default.range(start, end, step);
            var result = stream.toArray();
            chai_1.expect(result[0]).to.eq(start);
        });
        it('it should not include end value', function () {
            var start = 0;
            var end = 11;
            var step = 1;
            var stream = stream_1.default.range(start, end, step);
            var result = stream.toArray();
            chai_1.expect(result).to.not.contain.members([end]);
        });
        it('it should use step when provided', function () {
            var start = 2;
            var end = 10;
            var step = 2;
            var stream = stream_1.default.range(start, end, step);
            var result = stream.toArray();
            chai_1.expect(result[2] - result[1]).to.equal(step);
        });
        it('it should default step to 1 step when not provided', function () {
            var start = 2;
            var end = 10;
            var stream = stream_1.default.range(start, end);
            var result = stream.toArray();
            chai_1.expect(result[2] - result[1]).to.equal(1);
        });
        it('it should use absoulte step when given negative step', function () {
            var start = 2;
            var end = 10;
            var step = -2;
            var stream = stream_1.default.range(start, end, step);
            var result = stream.toArray();
            chai_1.expect(result[2] - result[1]).to.equal(Math.abs(step));
        });
        it('it should use negative step when start above end', function () {
            var start = 10;
            var end = 0;
            var step = 2;
            var stream = stream_1.default.range(start, end, step);
            var result = stream.toArray();
            chai_1.expect(result[2] - result[1]).to.equal(-step);
        });
        it('it should use negative step when start above end', function () {
            var start = 10;
            var end = 0;
            var stream = stream_1.default.range(start, end);
            var result = stream.toArray();
            chai_1.expect(result[2] - result[1]).to.equal(-1);
        });
        it('it should return empty if start and end are the same', function () {
            var start = 10;
            var end = 10;
            var stream = stream_1.default.range(start, end);
            var result = stream.toArray();
            chai_1.expect(result.length).to.be.eq(0);
        });
    });
    describe('rangeClosed', function () {
        it('it should start with start value', function () {
            var start = 0;
            var end = 11;
            var step = 1;
            var stream = stream_1.default.range(start, end, step);
            var result = stream.toArray();
            chai_1.expect(result[0]).to.eq(start);
        });
        it('it should include end value', function () {
            var start = 0;
            var end = 11;
            var step = 1;
            var stream = stream_1.default.rangeClosed(start, end, step);
            var result = stream.toArray();
            chai_1.expect(result).to.contain.members([end]);
        });
        it('it should use step when provided', function () {
            var start = 2;
            var end = 10;
            var step = 2;
            var stream = stream_1.default.rangeClosed(start, end, step);
            var result = stream.toArray();
            chai_1.expect(result[2] - result[1]).to.equal(step);
        });
        it('it should default step to 1 step when not provided', function () {
            var start = 2;
            var end = 10;
            var stream = stream_1.default.rangeClosed(start, end);
            var result = stream.toArray();
            chai_1.expect(result[2] - result[1]).to.equal(1);
        });
        it('it should use absoulte step when given negative step', function () {
            var start = 2;
            var end = 10;
            var step = -2;
            var stream = stream_1.default.rangeClosed(start, end, step);
            var result = stream.toArray();
            chai_1.expect(result[2] - result[1]).to.equal(Math.abs(step));
        });
        it('it should use negative step when start above end', function () {
            var start = 10;
            var end = 0;
            var step = 2;
            var stream = stream_1.default.rangeClosed(start, end, step);
            var result = stream.toArray();
            chai_1.expect(result[2] - result[1]).to.equal(-step);
        });
        it('it should use negative step when start above end', function () {
            var start = 10;
            var end = 0;
            var stream = stream_1.default.rangeClosed(start, end);
            var result = stream.toArray();
            chai_1.expect(result[2] - result[1]).to.equal(-1);
        });
        it('it should return a single value if start and end are the same', function () {
            var start = 10;
            var end = 10;
            var stream = stream_1.default.rangeClosed(start, end);
            var result = stream.toArray();
            chai_1.expect(result.length).to.be.eq(1);
        });
    });
});
